import logging
import os
import warnings
from ast import literal_eval
from pathlib import Path

import numpy as np
import pandas as pd
import xarray

from pyaerocom import const
from pyaerocom.exceptions import (
    CoordinateError,
    DataCoverageError,
    DataDimensionError,
    DataSourceError,
    MetaDataError,
    NetcdfError,
    UnknownRegion,
    VarNotAvailableError,
)
from pyaerocom.geodesy import get_country_info_coords
from pyaerocom.helpers import to_datestring_YYYYMMDD
from pyaerocom.helpers_landsea_masks import get_mask_value, load_region_mask_xr
from pyaerocom.mathutils import calc_statistics
from pyaerocom.plot.plotscatter import plot_scatter
from pyaerocom.region import Region
from pyaerocom.region_defs import REGION_DEFS
from pyaerocom.time_resampler import TimeResampler

logger = logging.getLogger(__name__)


class ColocatedData:
    """Class representing colocated and unified data from two sources

    Sources may be instances of :class:`UngriddedData` or
    :class:`GriddedData` that have been compared to each other.

    Note
    ----
    Currently, it is not foreseen, that this object is instantiated from
    scratch, but it is rather created in and returned by objects / methods
    that perform colocation.

    The purpose of this object is thus, not the creation of colocated objects,
    but solely the analysis of such data as well as I/O features (e.g. save as
    / read from .nc files, convert to pandas.DataFrame, plot station time
    series overlays, scatter plots, etc.).

    In the current design, such an object comprises 3 or 4 dimensions, where
    the first dimension (`data_source`, index 0) is ALWAYS length 2 and
    specifies the two datasets that were co-located (index 0 is obs, index 1
    is model). The second dimension is `time` and in case of 3D colocated data
    the 3rd dimension is `station_name` while for 4D colocated data the 3rd and
    4th dimension are latitude and longitude, respectively.

    3D colocated data is typically created when a model is colocated with
    station based ground based observations (
    cf :func:`pyaerocom.colocation.colocate_gridded_ungridded`) while 4D
    colocated data is created when a model is colocated with another model or
    satellite observations, that cover large parts of Earth's surface (other
    than discrete lat/lon pairs in the case of ground based station locations).

    Parameters
    ----------
    data : xarray.DataArray or numpy.ndarray or str, optional
        Colocated data. If str, then it is attempted to be loaded from file.
        Else, it is assumed that data is numpy array and that all further
        supplementary inputs (e.g. coords, dims) for the
        instantiation of :class:`DataArray` is provided via **kwargs.
    **kwargs
        Additional keyword args that are passed to init of :class:`DataArray`
        in case input `data` is numpy array.

    Raises
    ------
    IOError
        if init fails
    """

    __version__ = "0.11"

    def __init__(self, data=None, **kwargs):
        self._data = None
        if data is not None:
            if isinstance(data, Path):
                # make sure path is str instance
                data = str(data)
            if isinstance(data, str):
                self.open(data)
            elif isinstance(data, xarray.DataArray):
                self.data = data
            elif isinstance(data, np.ndarray):
                if not data.ndim in (3, 4):
                    raise DataDimensionError("invalid input, need 3D or 4D numpy array")
                elif not data.shape[0] == 2:
                    raise DataDimensionError(
                        "first dimension (data_source) must be of length 2(obs, model)"
                    )
                data = xarray.DataArray(data, **kwargs)
                self.data = data
            else:
                raise ValueError(f"Failed to interpret input {data}")

    @property
    def data(self):
        """:class:`xarray.DataArray` containing colocated data

        Raises
        ------
        AttributeError
            if data is not available

        Returns
        -------
        xarray.DataArray
            array containing colocated data and metadata (in fact, there is no
            additional attributes to `ColocatedData` and everything is contained
            in :attr:`data`).
        """
        if self._data is None:
            raise AttributeError("No data available in this object")
        return self._data

    @data.setter
    def data(self, val):
        if not isinstance(val, xarray.DataArray):
            raise ValueError("Invalid input for data attribute, need instance of xarray.DataArray")
        self._data = val

    @property
    def ndim(self):
        """Dimension of data array"""
        return self.data.ndim

    @property
    def coords(self):
        """Coordinates of data array"""
        return self.data.coords

    @property
    def dims(self):
        """Names of dimensions"""
        return self.data.dims

    @property
    def shape(self):
        """Shape of data array"""
        return self.data.shape

    @property
    def data_source(self):
        """Coordinate array containing data sources (z-axis)"""
        return self.data.data_source

    @property
    def model_name(self):
        if "model_name" in self.metadata:
            return self.metadata["model_name"]
        return self.data_source[1]

    @property
    def obs_name(self):
        if "obs_name" in self.metadata:
            return self.metadata["obs_name"]
        return self.data_source[0]

    @property
    def var_name(self):
        """Coordinate array containing data sources (z-axis)"""
        return self.data.var_name

    @property
    def longitude(self):
        """Array of longitude coordinates"""
        if not "longitude" in self.data.coords:
            raise AttributeError("ColocatedData does not include longitude coordinate")
        return self.data.longitude

    @property
    def lon_range(self):
        """Longitude range covered by this data object"""
        lons = self.longitude.values
        return (np.nanmin(lons), np.nanmax(lons))

    @property
    def latitude(self):
        """Array of latitude coordinates"""
        if not "latitude" in self.data.coords:
            raise AttributeError("ColocatedData does not include latitude coordinate")
        return self.data.latitude

    @property
    def lat_range(self):
        """Latitude range covered by this data object"""
        lats = self.latitude.values
        return (np.nanmin(lats), np.nanmax(lats))

    @property
    def time(self):
        """Array containing time stamps"""
        if not "time" in self.data.dims:
            raise AttributeError("ColocatedData does not include time coordinate")
        return self.data.time

    @property
    def start(self):
        """Start datetime of data"""
        return self.time.values[0]

    @property
    def stop(self):
        """Stop datetime of data"""
        return self.time.values[-1]

    @property
    def ts_type(self):
        """String specifying temporal resolution of data"""
        if not "ts_type" in self.metadata:
            raise ValueError(
                "Colocated data object does not contain information about temporal resolution"
            )
        return self.metadata["ts_type"]

    @property
    def start_str(self):
        """
        str: Start date of data as str with format YYYYMMDD
        """
        return to_datestring_YYYYMMDD(self.start)

    @property
    def stop_str(self):
        """
        str: Stop date of data as str with format YYYYMMDD
        """
        return to_datestring_YYYYMMDD(self.stop)

    @property
    def units(self):
        """Unit of data"""
        return self.data.attrs["var_units"]

    @property
    def unitstr(self):
        """String representation of obs and model units in this object"""
        unique = []
        u = self.units
        for val in u:
            if val is None:
                val = "N/D"
            elif not isinstance(val, str):
                val = str(val)
            if not val in unique:
                unique.append(val)
        return ", ".join(unique)

    @property
    def metadata(self):
        """Meta data dictionary (wrapper to :attr:`data.attrs`"""
        return self.data.attrs

    @property
    def num_coords(self):
        """Total number of lat/lon coordinate pairs"""
        obj = self.flatten_latlondim_station_name() if self.has_latlon_dims else self
        if not "station_name" in obj.coords:
            raise DataDimensionError("Need dimension station_name")
        return len(obj.data.station_name)

    @property
    def num_coords_with_data(self):
        """Number of lat/lon coordinate pairs that contain at least one datapoint

        Note
        ----
        Occurrence of valid data is only checked for obsdata (first index in
        data_source dimension).
        """
        obj = self.flatten_latlondim_station_name() if self.has_latlon_dims else self
        dims = obj.dims
        if not "station_name" in dims:
            raise DataDimensionError("Need dimension station_name")
        obs = obj.data[0]
        if len(dims) > 3:  # additional dimensions
            default_dims = ("data_source", "time", "station_name")
            add_dims = tuple(x for x in dims if not x in default_dims)
            raise DataDimensionError(
                f"Can only unambiguously retrieve no of coords with obs data "
                f"for colocated data with dims {default_dims}, please reduce "
                f"dimensionality first by selecting or aggregating additional "
                f"dimensions: {add_dims}"
            )

        if "time" in dims:
            val = (obs.count(dim="time") > 0).data.sum()
        else:
            val = (~np.isnan(obs.data)).sum()
        return val

    @property
    def has_time_dim(self):
        """Boolean specifying whether data has a time dimension"""
        return True if "time" in self.dims else False

    @property
    def has_latlon_dims(self):
        """Boolean specifying whether data has latitude and longitude dimensions"""
        return all([dim in self.dims for dim in ["latitude", "longitude"]])

    @property
    def countries_available(self):
        """
        Alphabetically sorted list of country names available

        Raises
        ------
        MetaDataError
            if no country information is available

        Returns
        -------
        list
            list of countries available in these data
        """
        if not "country" in self.coords:
            raise MetaDataError(
                "No country information available in "
                "ColocatedData. You may run class method "
                "check_set_countries to automatically assign "
                "countries to the station_name coordinate"
            )
        return sorted(dict.fromkeys(self.data["country"].data))

    @property
    def country_codes_available(self):
        """
        Alphabetically sorted list of country codes available

        Raises
        ------
        MetaDataError
            if no country information is available

        Returns
        -------
        list
            list of countries available in these data
        """
        if not "country_code" in self.coords:
            raise MetaDataError(
                "No country information available in "
                "ColocatedData. You may run class method "
                "check_set_countries to automatically assign "
                "countries to the station_name coordinate"
            )
        return sorted(dict.fromkeys(self.data["country_code"].data))

    @property
    def area_weights(self):
        """
        Wrapper for :func:`calc_area_weights`
        """
        return self.calc_area_weights()

    def get_meta_item(self, key: str):
        """
        Get metadata value

        Parameters
        ----------
        key : str
            meta item key.

        Raises
        ------
        AttributeError
            If key is not available.

        Returns
        -------
        object
            value of metadata.

        """

        try:
            return self.metadata[key]
        except KeyError:
            raise AttributeError(f"no such meta-key available {key}")

    def get_country_codes(self):
        """
        Get country names and codes for all locations contained in these data

        Raises
        ------
        MetaDataError
            if no country information is available

        Returns
        -------
        dict
            dictionary of unique country names (keys) and corresponding country
            codes (values)
        """
        if not "country" in self.coords:
            raise MetaDataError(
                "No country information available in "
                "ColocatedData. You may run class method "
                "check_set_countries to automatically assign "
                "countries to the station_name coordinate"
            )
        countries = self.data["country"].data
        codes = self.data["country_code"].data
        return dict(zip(countries, codes))

    def calc_area_weights(self):
        """Calculate area weights

        Note
        ----
        Only applies to colocated data that has latitude and longitude
        dimension.

        Returns
        -------
        ndarray
            array containing weights for each datapoint (same shape as
            `self.data[0]`)
        """
        if not self.has_latlon_dims:
            raise DataDimensionError(
                "Can only compute area weights for data with latitude and longitude dimension"
            )
        if not "units" in self.data.latitude.attrs:
            self.data.latitude.attrs["units"] = "degrees"
        if not "units" in self.data.longitude.attrs:
            self.data.longitude.attrs["units"] = "degrees"
        arr = self.data
        from pyaerocom import GriddedData

        obs = GriddedData(arr.to_iris(), check_unit=False, convert_unit_on_init=False)
        return obs.calc_area_weights()

    def min(self):
        """
        Wrapper for :func:`xarray.DataArray.min` called from :attr:`data`

        Returns
        -------
        xarray.DataArray
            minimum of data

        """
        return self.data.min()

    def max(self):
        """
        Wrapper for :func:`xarray.DataArray.max` called from :attr:`data`

        Returns
        -------
        xarray.DataArray
            maximum of data

        """
        return self.data.max()

    def resample_time(
        self,
        to_ts_type,
        how=None,
        min_num_obs=None,
        colocate_time=False,
        settings_from_meta=False,
        inplace=False,
        **kwargs,
    ):
        """
        Resample time dimension

        The temporal resampling is done using :class:`TimeResampler`

        Parameters
        ----------
        to_ts_type : str
            desired output frequency.
        how : str or dict, optional
            aggregator used for resampling (e.g. max, min, mean, median). Can
            also be hierarchical scheme via `dict`, similar to `min_num_obs`.
            The default is None.
        min_num_obs : int or dict, optional
            Minimum number of observations required to resample from current
            frequency (:attr:`ts_type`) to desired output frequency.
        colocate_time : bool, optional
            If True, the modeldata is invalidated where obs is NaN, before
            resampling. The default is False (updated in v0.11.0, before was
            True).
        settings_from_meta : bool
            if True, then input args `how`, `min_num_obs` and `colocate_time`
            are ignored and instead the corresponding values set in
            :attr:`metadata` are used. Defaults to False.
        inplace : bool, optional
            If True, modify this object directly, else make a copy and resample
            that one. The default is False (updated in v0.11.0, before was
            True).
        **kwargs
            Addtitional keyword args passed to :func:`TimeResampler.resample`.

        Returns
        -------
        ColocatedData
            Resampled colocated data object.

        """
        if inplace:
            col = self
        else:
            col = self.copy()

        if settings_from_meta:
            how = self.metadata["resample_how"]
            min_num_obs = self.metadata["min_num_obs"]
            colocate_time = self.metadata["colocate_time"]

        # if colocate time is activated, remove datapoints from model, where
        # there is no observation
        if colocate_time:
            mask = np.isnan(col.data[0]).data
            col.data.data[1][mask] = np.nan

        res = TimeResampler(col.data)
        data_arr = res.resample(
            to_ts_type=to_ts_type,
            from_ts_type=col.ts_type,
            how=how,
            min_num_obs=min_num_obs,
            **kwargs,
        )

        data_arr.attrs.update(col.metadata)
        data_arr.attrs["ts_type"] = str(to_ts_type)

        col.data = data_arr
        col.data.attrs["colocate_time"] = colocate_time
        trs = res.last_setup
        trs["resample_how"] = trs.pop("how")
        col.data.attrs.update(trs)
        return col

    def flatten_latlondim_station_name(self):
        """Stack (flatten) lat / lon dimension into new dimension station_name

        Returns
        -------
        ColocatedData
            new colocated data object with dimension station_name and lat lon
            arrays as additional coordinates
        """
        if not self.has_latlon_dims:
            raise DataDimensionError("Need latitude and longitude dimensions")

        newdims = []
        for dim in self.dims:
            if dim == "latitude":
                newdims.append("station_name")
            elif dim == "longitude":
                continue
            else:
                newdims.append(dim)

        arr = self.stack(station_name=["latitude", "longitude"], inplace=False).data

        arr = arr.transpose(*newdims)
        return ColocatedData(arr)

    def stack(self, inplace=False, **kwargs):
        """Stack one or more dimensions

        For details see :func:`xarray.DataArray.stack`.

        Parameters
        ----------
        inplace : bool
            modify this object or a copy.
        **kwargs
            input arguments passed to :func:`DataArray.stack`

        Returns
        -------
        ColocatedData
            stacked data object
        """
        if inplace:
            data = self
        else:
            data = self.copy()
        data.data = data.data.stack(**kwargs)
        return data

    def unstack(self, inplace=False, **kwargs):
        """Unstack one or more dimensions

        For details see :func:`xarray.DataArray.unstack`.

        Parameters
        ----------
        inplace : bool
            modify this object or a copy.
        **kwargs
            input arguments passed to :func:`DataArray.unstack`

        Returns
        -------
        ColocatedData
            unstacked data object
        """
        if inplace:
            data = self
        else:
            data = self.copy()
        return data.unstack(**kwargs)

    def get_coords_valid_obs(self):
        """
        Get latitude / longitude coordinates where obsdata is available

        Returns
        -------
        list
            latitute coordinates
        list
            longitude coordinates

        """

        obs = self.data[0]
        if self.ndim == 4:
            stacked = obs.stack(x=["latitude", "longitude"])
            invalid = stacked.isnull().all(dim="time")
            coords = stacked.x[~invalid].values
            coords = zip(*list(coords))
        else:
            invalid = obs.isnull().all(dim="time")
            lats = list(obs.latitude[~invalid].values)
            lons = list(obs.longitude[~invalid].values)
            coords = (lats, lons)
        return list(coords)

    def _iter_stats(self):
        """Create a list that can be used to iterate over station dimension

        Returns
        -------
        list
            list containing 3-element tuples, one for each site i, comprising
            (latitude[i], longitude[i], station_name[i]).
        """
        if not "station_name" in self.data.dims:
            raise AttributeError(
                "ColocatedData object has no dimension station_name. Consider stacking..."
            )
        if "latitude" in self.dims and "longitude" in self.dims:
            raise AttributeError(
                "Cannot init station iter index since latitude and longitude are othorgonal"
            )
        lats = self.data.latitude.values
        lons = self.data.longitude.values
        stats = self.data.station_name.values

        return list(zip(lats, lons, stats))

    def _get_stat_coords(self):
        """
        Get station coordinates

        Raises
        ------
        DataDimensionError
            if data is 4D and does not have latitude and longitude dimension

        Returns
        -------
        list
            list containing 2 element tuples (latitude, longitude)

        """
        if self.ndim == 4:
            if not self.has_latlon_dims:
                raise DataDimensionError("Invalid dimensions in 4D ColocatedData")
            lats, lons = self.data.latitude.data, self.data.longitude.data
            coords = np.dstack(np.meshgrid(lats, lons))
            coords = coords.reshape(len(lats) * len(lons), 2)
        else:
            coords = zip(self.latitude.data, self.longitude.data)
        return list(coords)

    def check_set_countries(self, inplace=True, assign_to_dim=None):
        """
        Checks if country information is available and assigns if not

        If not country information is available, countries will be assigned
        for each lat / lon coordinate using
        :func:`pyaerocom.geodesy.get_country_info_coords`.

        Parameters
        ----------
        inplace : bool, optional
            If True, modify and return this object, else a copy.
            The default is True.
        assign_to_dim : str, optional
            name of dimension to which the country coordinate is assigned.
            Default is None, in which case station_name is used.

        Raises
        ------
        DataDimensionError
            If data is 4D (i.e. if latitude and longitude are othorgonal
            dimensions)

        Returns
        -------
        ColocatedData
            data object with countries assigned

        """
        if self.has_latlon_dims:
            raise DataDimensionError(
                "Countries cannot be assigned to 4D"
                "ColocatedData with othorgonal lat / lon "
                "dimensions. Please consider stacking "
                "the latitude and longitude dimensions-"
            )
        if assign_to_dim is None:
            assign_to_dim = "station_name"

        if not assign_to_dim in self.dims:
            raise DataDimensionError("No such dimension", assign_to_dim)

        coldata = self if inplace else self.copy()

        if "country" in coldata.data.coords:
            logger.info("Country information is available")
            return coldata
        coords = coldata._get_stat_coords()

        info = get_country_info_coords(coords)

        countries, codes = [], []
        for item in info:
            countries.append(item["country"])
            codes.append(item["country_code"])

        arr = coldata.data
        arr = arr.assign_coords(
            country=(assign_to_dim, countries), country_code=(assign_to_dim, codes)
        )
        coldata.data = arr
        return coldata

    def copy(self):
        """Copy this object"""
        return ColocatedData(self.data.copy())

    def set_zeros_nan(self, inplace=True):
        """
        Replace all 0's with NaN in data

        Parameters
        ----------
        inplace : bool
            Whether to modify this object or a copy. The default is True.

        Returns
        -------
        cd : ColocatedData
            modified data object

        """
        cd = self if inplace else self.copy()
        mask = cd.data.data == 0
        cd.data.data[mask] = np.nan
        cd.metadata["zeros_to_nan"] = True

        return cd

    def calc_statistics(self, use_area_weights=False, **kwargs):
        """Calculate statistics from model and obs data

        Calculate standard statistics for model assessment. This is done by
        taking all model and obs data points in this object as input for
        :func:`pyaerocom.mathutils.calc_statistics`. For instance, if the
        object is 3D with dimensions `data_source` (obs, model), `time` (e.g.
        12 monthly values) and `station_name` (e.g. 4 sites), then the input
        arrays for model and obs into
        :func:`pyaerocom.mathutils.calc_statistics` will be each of size
        12x4.

        See also :func:`calc_temporal_statistics` and
        :func:`calc_spatial_statistics`.

        Parameters
        ----------
        use_area_weights : bool
            if True and if data is 4D (i.e. has lat and lon dimension), then
            area weights are applied when caluclating the statistics based on
            the coordinate cell sizes. Defaults to False.
        **kwargs
            additional keyword args passed to
            :func:`pyaerocom.mathutils.calc_statistics`

        Returns
        -------
        dict
            dictionary containing statistical parameters
        """
        if use_area_weights and not "weights" in kwargs and self.has_latlon_dims:
            kwargs["weights"] = self.area_weights[0].flatten()

        nc = self.num_coords
        try:
            ncd = self.num_coords_with_data
        except DataDimensionError:
            ncd = np.nan
        obsvals = self.data.values[0].flatten()
        modvals = self.data.values[1].flatten()
        stats = calc_statistics(modvals, obsvals, **kwargs)

        stats["num_coords_tot"] = nc
        stats["num_coords_with_data"] = ncd
        return stats

    def calc_temporal_statistics(self, aggr=None, **kwargs):
        """Calculate *temporal* statistics from model and obs data

        *Temporal* statistics is computed by averaging first the spatial
        dimension(s) (that is, `station_name` for 3D data, and
        `latitude` and `longitude` for 4D data), so that only `data_source` and
        `time` remains as dimensions. These 2D data are then used to calculate
        standard statistics using :func:`pyaerocom.mathutils.calc_statistics`.

        See also :func:`calc_statistics` and
        :func:`calc_spatial_statistics`.

        Parameters
        ----------
        aggr : str, optional
            aggreagator to be used, currently only mean and median are
            supported. Defaults to mean.
        **kwargs
            additional keyword args passed to
            :func:`pyaerocom.mathutils.calc_statistics`

        Returns
        -------
        dict
            dictionary containing statistical parameters
        """
        if aggr is None:
            aggr = "mean"
        nc = self.num_coords
        try:
            ncd = self.num_coords_with_data
        except:
            ncd = np.nan
        if self.has_latlon_dims:
            dim = ("latitude", "longitude")
        else:
            dim = "station_name"

        if aggr == "mean":
            arr = self.data.mean(dim=dim)
        elif aggr == "median":
            arr = self.data.median(dim=dim)
        else:
            raise ValueError("So far only mean and median are supported aggregators")
        obs, mod = arr[0].values.flatten(), arr[1].values.flatten()
        stats = calc_statistics(mod, obs, **kwargs)
        stats["num_coords_tot"] = nc
        stats["num_coords_with_data"] = ncd
        return stats

    def calc_spatial_statistics(self, aggr=None, use_area_weights=False, **kwargs):
        """Calculate *spatial* statistics from model and obs data

        *Spatial* statistics is computed by averaging first the time
        dimension and then, if data is 4D, flattening lat / lon dimensions into
        new station_name dimension, so that the resulting dimensions are
        `data_source` and `station_name`. These 2D data are then used to
        calculate standard statistics using
        :func:`pyaerocom.mathutils.calc_statistics`.

        See also :func:`calc_statistics` and
        :func:`calc_temporal_statistics`.

        Parameters
        ----------
        aggr : str, optional
            aggreagator to be used, currently only mean and median are
            supported. Defaults to mean.
        use_area_weights : bool
            if True and if data is 4D (i.e. has lat and lon dimension), then
            area weights are applied when caluclating the statistics based on
            the coordinate cell sizes. Defaults to False.
        **kwargs
            additional keyword args passed to
            :func:`pyaerocom.mathutils.calc_statistics`

        Returns
        -------
        dict
            dictionary containing statistical parameters
        """
        if aggr is None:
            aggr = "mean"
        if use_area_weights and not "weights" in kwargs and self.has_latlon_dims:
            weights = self.area_weights[0]  # 3D (time, lat, lon)
            assert self.dims[1] == "time"
            kwargs["weights"] = np.nanmean(weights, axis=0).flatten()

        nc, ncd = self.num_coords, self.num_coords_with_data
        # ToDo: find better solution to parse aggregator without if conditions,
        # e.g. xr.apply_ufunc or similar, with core aggregators that are
        # supported being defined in some dictionary in some pyaerocom config
        # module or class. Also aggregation could go into a separate method...
        if aggr == "mean":
            arr = self.data.mean(dim="time")
        elif aggr == "median":
            arr = self.data.median(dim="time")
        else:
            raise ValueError("So far only mean and median are supported aggregators")

        obs, mod = arr[0].values.flatten(), arr[1].values.flatten()
        stats = calc_statistics(mod, obs, **kwargs)
        stats["num_coords_tot"] = nc
        stats["num_coords_with_data"] = ncd
        return stats

    def plot_scatter(self, **kwargs):
        """Create scatter plot of data

        Parameters
        ----------
        **kwargs
            keyword args passed to :func:`pyaerocom.plot.plotscatter.plot_scatter`

        Returns
        -------
        Axes
            matplotlib axes instance
        """
        meta = self.metadata
        try:
            num_points = self.num_coords_with_data
        except DataDimensionError:
            num_points = np.nan
        try:
            vars_ = meta["var_name"]
        except KeyError:
            vars_ = ["N/D", "N/D"]
        try:
            xn, yn = meta["data_source"]
        except KeyError:
            xn, yn = "N/D", "N/D"

        if vars_[0] != vars_[1]:
            var_ref = vars_[0]
        else:
            var_ref = None
        try:
            tst = meta["ts_type"]
        except KeyError:
            tst = "N/D"
        try:
            fn = meta["filter_name"]
        except KeyError:
            fn = "N/D"
        try:
            unit = self.unitstr
        except KeyError:
            unit = "N/D"
        try:
            start = self.start
        except AttributeError:
            start = "N/D"

        try:
            stop = self.stop
        except AttributeError:
            stop = "N/D"

        # ToDo: include option to use area weighted stats in plotting
        # routine...
        return plot_scatter(
            x_vals=self.data.values[0].flatten(),
            y_vals=self.data.values[1].flatten(),
            var_name=vars_[1],
            var_name_ref=var_ref,
            x_name=xn,
            y_name=yn,
            start=start,
            stop=stop,
            unit=unit,
            ts_type=tst,
            stations_ok=num_points,
            filter_name=fn,
            **kwargs,
        )

    def rename_variable(self, var_name, new_var_name, data_source, inplace=True):
        """Rename a variable in this object

        Parameters
        ----------
        var_name : str
            current variable name
        new_var_name : str
            new variable name
        data_source : str
            name of data source (along data_source dimension)
        inplace : bool
            replace here or create new instance

        Returns
        -------
        ColocatedData
            instance with renamed variable

        Raises
        ------
        VarNotAvailableError
            if input variable is not available in this object
        DataSourceError
            if input data_source is not available in this object
        """
        if not data_source in self.metadata["data_source"]:
            raise DataSourceError(f"No such data source {data_source} in ColocatedData")
        if not var_name in self.metadata["var_name"]:
            raise VarNotAvailableError(f"No such variable {var_name} in ColocatedData")

        if inplace:
            obj = self
        else:
            obj = self.copy()
        arr = obj.data
        idx = arr.attrs["data_source"].index(data_source)
        arr.attrs["var_name_orig"] = arr.attrs["var_name"]
        arr.attrs["var_name"][idx] = new_var_name

        if var_name == arr.name:
            arr.name = new_var_name
        obj.data = arr
        return obj

    @staticmethod
    def _aerocom_savename(
        obs_var, obs_id, mod_var, mod_id, start_str, stop_str, ts_type, filter_name
    ):
        return (
            f"{mod_var}_{obs_var}_MOD-{mod_id}_REF-{obs_id}_"
            f"{start_str}_{stop_str}_{ts_type}_{filter_name}"
        )

    @property
    def savename_aerocom(self):
        """Default save name for data object following AeroCom convention"""
        obsid, modid = self.metadata["data_source"]
        obsvar, modvar = self.metadata["var_name"]

        return self._aerocom_savename(
            obsvar,
            obsid,
            modvar,
            modid,
            self.start_str,
            self.stop_str,
            self.ts_type,
            self.metadata["filter_name"],
        )

    @staticmethod
    def get_meta_from_filename(file_path):
        """Get meta information from file name

        Note
        ----
        This does not yet include IDs of model and obs data as these should
        be included in the data anyways (e.g. column names in CSV file) and
        may include the delimiter _ in their name.

        Returns
        -------
        dict
            dicitonary with meta information
        """

        spl = os.path.basename(file_path).split(".nc")[0].split("_")

        if not spl[2].startswith("MOD"):
            raise ValueError("File name does not follow convention")

        modname = spl[2].split("MOD-")[1]
        refname = ""
        in_ref = False
        for item in spl[3:-4]:
            if in_ref:
                refname += f"_{item}"
            elif item.startswith("REF-"):
                in_ref = True
                refname = item.split("REF-")[1]
            elif not in_ref:
                modname += f"_{item}"

        meta = dict(
            model_var=spl[0],
            obs_var=spl[1],
            model_name=modname,
            obs_name=refname,
            start=spl[-4],
            stop=spl[-3],
            ts_type=spl[-2],
            filter_name=spl[-1],
        )

        return meta

    def get_time_resampling_settings(self):
        """Returns a dictionary with relevant settings for temporal resampling

        Returns
        -------
        dict
        """
        settings = {}
        mapping = {
            "min_num_obs": "min_num_obs",
            "resample_how": "resample_how",
            "colocate_time": "colocate_time",
        }
        for from_key, to_key in mapping.items():
            try:
                settings[to_key] = self.metadata[from_key]
            except KeyError:
                logger.warning(f"Meta key {from_key} not defined in ColocatedData.meta...")
                settings[to_key] = None
        return settings

    def _prepare_meta_to_netcdf(self):
        """
        Prepare metada for NetCDF format

        Returns
        -------
        meta_out : dict
            metadata ready for serialisation to NetCDF.

        """
        meta = self.data.attrs
        meta_out = {}
        for key, val in meta.items():
            if val is None:
                meta_out[key] = "None"
            elif isinstance(val, bool):
                meta_out[key] = int(val)
            elif isinstance(val, dict):
                meta_out[f"CONV!{key}"] = str(val)
            else:
                meta_out[key] = val
        return meta_out

    def to_netcdf(self, out_dir, savename=None, **kwargs):
        """Save data object as NetCDF file

        Wrapper for method :func:`xarray.DataArray.to_netdcf`

        Parameters
        ----------
        out_dir : str
            output directory
        savename : str, optional
            name of file, if None, the default save name is used (cf.
            :attr:`savename_aerocom`)
        **kwargs
            additional, optional keyword arguments passed to
            :func:`xarray.DataArray.to_netdcf`

        Returns
        -------
        str
            file path of stored object.
        """
        if "path" in kwargs:
            raise OSError("Path needs to be specified using input parameters out_dir and savename")
        if savename is None:
            savename = self.savename_aerocom
        if not savename.endswith(".nc"):
            savename = f"{savename}.nc"
        arr = self.data.copy()
        arr.attrs = self._prepare_meta_to_netcdf()
        fp = os.path.join(out_dir, savename)
        arr.to_netcdf(path=fp, **kwargs)
        return fp

    def _meta_from_netcdf(self, imported_meta):
        """
        Convert imported metadata as stored in NetCDF file

        Reason for this is that some meta values cannot be serialised when
        storing as NetCDF (e.g. nested dictionary `min_num_obs` or None values)

        Parameters
        ----------
        imported_meta : dict
            metadata as read in from colocated NetCDF file

        Returns
        -------
        meta : dict
            converted meta

        """
        meta = {}
        for key, val in imported_meta.items():
            if isinstance(val, str) and val == "None":
                meta[key] = None
            elif key.startswith("CONV!"):
                key = key[5:]
                meta[key] = literal_eval(val)
            else:
                meta[key] = val
        return meta

    def read_netcdf(self, file_path):
        """Read data from NetCDF file

        Parameters
        ----------
        file_path : str
            file path
        """
        try:
            self.get_meta_from_filename(file_path)
        except Exception as e:
            raise NetcdfError(
                f"Invalid file name for ColocatedData: {file_path}. Error: {repr(e)}"
            )

        arr = xarray.open_dataarray(file_path)
        arr.attrs = self._meta_from_netcdf(arr.attrs)
        self.data = arr
        return self

    def to_dataframe(self):
        """Convert this object into pandas.DataFrame

        Note
        ----
        This does not include meta information
        """
        logger.warning("This method is currently not completely finished")
        model_vals = self.data.values[1].flatten()
        obs_vals = self.data.values[0].flatten()
        mask = ~np.isnan(obs_vals)
        return pd.DataFrame({"ref": obs_vals[mask], "data": model_vals[mask]})

    def from_dataframe(self, df):
        """Create colocated Data object from dataframe

        Note
        ----
        This is intended to be used as back-conversion from :func:`to_dataframe`
        and methods that use the latter (e.g. :func:`to_csv`).
        """
        raise NotImplementedError("Coming soon...")
        data = df.to_xarray()
        self.data = data

    def to_csv(self, out_dir, savename=None):
        """Save data object as .csv file

        Converts data to pandas.DataFrame and then saves as csv

        Parameters
        ----------
        out_dir : str
            output directory
        savename : :obj:`str`, optional
            name of file, if None, the default save name is used (cf.
            :attr:`savename_aerocom`)
        """
        if savename is None:
            savename = self.savename_aerocom
        if not savename.endswith(".csv"):
            savename = f"{savename}.csv"
        df = self.to_dataframe()
        file_path = os.path.join(out_dir, savename)
        df.to_csv(file_path)
        return file_path

    def from_csv(self, file_path):
        """Read data from CSV file

        Todo
        -----
        Complete docstring
        """
        meta = self.get_meta_from_filename(file_path)
        df = pd.read_csv(file_path)
        self.from_dataframe(df)
        self.data.attrs.update(**meta)

    def open(self, file_path):
        """High level helper for reading from supported file sources

        Parameters
        ----------
        file_path : str
            file path
        """
        if file_path.endswith("nc"):
            self.read_netcdf(file_path)
            return

        raise OSError(
            f"Failed to import file {os.path.basename(file_path)}. File type is not supported "
        )

    def filter_altitude(self, alt_range, inplace=False):
        """
        Apply altitude filter

        Parameters
        ----------
        alt_range : list or tuple
            altitude range to be applied to data (2-element list)
        inplace : bool, optional
            Apply filter to this object directly or to a copy.
            The default is False.

        Raises
        ------
        NotImplementedError
            If data is 4D, i.e. it contains latitude and longitude dimensions.

        Returns
        -------
        ColocatedData
            Filtered data object .

        """
        if not self._check_latlon_coords():
            raise NotImplementedError(
                "Altitude filtering for data with lat and lon dimension is not yet supported"
            )
        filtered = self._filter_altitude_2d(self.data, alt_range)
        filtered.attrs["alt_range"] = alt_range
        if inplace:
            self.data = filtered
            return self
        return ColocatedData(filtered)

    @staticmethod
    def _filter_altitude_2d(arr, alt_range):
        """
        Apply altitude filter to 2D spatiotemporal data

        Parameters
        ----------
        arr : DataArray
            input array to be filtered
        alt_range : list or tuple
            altitude range to be applied to data (2-element list)

        Raises
        ------
        DataDimensionError
            if station_name is not the 3rd index

        Returns
        -------
        DataArray
            Filtered object.

        """
        if not list(arr.dims).index("station_name") == 2:
            raise DataDimensionError("station_name must be 3. dimensional index")

        mask = np.logical_and(arr.altitude > alt_range[0], arr.altitude < alt_range[1])

        filtered = arr[:, :, mask]
        return filtered

    def _check_latlon_coords(self):
        """
        Check if latitude and longitude coordinates exists and are NOT dimensions

        Raises
        ------
        CoordinateError
            - if any or both coordinates are missing.
            - if one (and only one) of both coords is also a dimension

        Returns
        -------
        bool
            True if latitude and longitude are existing coordinates, False if
            they are also dimensions.

        """
        _check = ("latitude", "longitude")
        if not all([x in self.coords for x in _check]):
            raise CoordinateError("Missing latitude or longitude coordinate (or both)")
        elif any([x in self.dims for x in _check]):
            if not all([x in self.dims for x in _check]):
                raise CoordinateError(
                    "Only one of latitude / longitude is dimension (require None or both)"
                )
            return False
        return True

    @staticmethod
    def _filter_country_2d(arr, country, use_country_code):
        """
        Apply country filter in 2D spatiotemporal data

        Parameters
        ----------
        arr : DataArray
            data to be filtered for country.
        country : str
            name of country.
        use_country_code : bool
            If True, input value for `country` is evaluated against country
            codes rather than country names.

        Raises
        ------
        DataDimensionError
            If `country` is not a coordinate of input `DataArray`, or, in case
            `use_country_code=True`, if `country_code` is not a coordinate.
        DataCoverageError
            if filtering results in empty data object.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        what = "country" if not use_country_code else "country_code"
        if not what in arr.coords:
            raise DataDimensionError(
                f"Cannot filter country {country}. No country information available in DataArray"
            )
        mask = arr[what] == country
        if mask.sum() == 0:
            raise DataCoverageError(f"No data available in country {country} in ColocatedData")
        return arr[:, :, mask]

    @staticmethod
    def _filter_latlon_2d(arr, lat_range, lon_range):
        """
        Apply rectangular latitude / longitude filter for 2D data

        2D here means in terms of spatiotemporal dimensions, that is, the
        object has to have one spatial dimension `station_name` and one
        temporal dimension `time`.

        ToDo
        ----
        Use `sel` with slice, or similar, need to check though how that will
        apply to filtering that crosses the +180 -> -180 degree longitude
        border.

        Parameters
        ----------
        arr : DataArray
            data to be filtered.
        lat_range : list, optional
            latitude range that is supposed to be applied. If specified, then
            also lon_range need to be specified, else, region_id is checked
            against AeroCom default regions (and used if applicable)
        lon_range : list, optional
            longitude range that is supposed to be applied. If specified, then
            also lat_range need to be specified, else, region_id is checked
            against AeroCom default regions (and used if applicable)

        Raises
        ------
        DataDimensionError
            If index of `station_name` dimension is not at 3rd position in the
            underlying array.
        DataCoverageError
            If no data is available in the specified rectangular region.

        Returns
        -------
        ColocatedData
            Filtered data object.

        """
        if not list(arr.dims).index("station_name") == 2:
            raise DataDimensionError("station_name dimension must be at 3rd index position")

        lons, lats = arr.longitude.data, arr.latitude.data

        latmask = np.logical_and(lats > lat_range[0], lats < lat_range[1])
        if lon_range[0] > lon_range[1]:
            _either = np.logical_and(lons >= -180, lons < lon_range[1])
            _or = np.logical_and(lons > lon_range[0], lons <= 180)
            lonmask = np.logical_or(_either, _or)
        else:
            lonmask = np.logical_and(lons > lon_range[0], lons < lon_range[1])
        mask = latmask & lonmask
        if mask.sum() == 0:
            raise DataCoverageError(
                f"No data available in latrange={lat_range} and "
                f"lonrange={lon_range} in ColocatedData"
            )

        return arr[:, :, mask]

    @staticmethod
    def _filter_latlon_3d(arr, lat_range, lon_range):
        """
        Apply rectangular latitude / longitude filter for 3D data

        32D here means in terms of spatiotemporal dimensions, that is, the
        object has to have two spatial dimensions `latitude` and `longitude`
        and one temporal dimension `time`.

        ToDo
        ----
        Implement filtering that crosses the +180 -> -180 degree longitude
        border.

        Parameters
        ----------
        arr : DataArray
            data to be filtered.
        lat_range : list, optional
            latitude range that is supposed to be applied. If specified, then
            also lon_range need to be specified, else, region_id is checked
            against AeroCom default regions (and used if applicable)
        lon_range : list, optional
            longitude range that is supposed to be applied. If specified, then
            also lat_range need to be specified, else, region_id is checked
            against AeroCom default regions (and used if applicable)

        Raises
        ------
        NotImplementedError
            If input longitude bounds cross the +180 -> -180 border.

        Returns
        -------
        ColocatedData
            Filtered data object.
        """
        if lon_range[0] > lon_range[1]:
            raise NotImplementedError(
                "Filtering longitude over 180 deg edge is not yet possible in "
                "3D ColocatedData..."
            )
        if not isinstance(lat_range, slice):
            lat_range = slice(lat_range[0], lat_range[1])
        if not isinstance(lon_range, slice):
            lon_range = slice(lon_range[0], lon_range[1])

        return arr.sel(dict(latitude=lat_range, longitude=lon_range))

    def apply_country_filter(self, region_id, use_country_code=False, inplace=False):
        """
        Apply country filter

        Parameters
        ----------
        region_id : str
            country name or code.
        use_country_code : bool, optional
            If True, input value for `country` is evaluated against country
            codes rather than country names. Defaults to False.
        inplace : bool, optional
            Apply filter to this object directly or to a copy.
            The default is False.

        Raises
        ------
        NotImplementedError
            if data is 4D (i.e. it has latitude and longitude dimensions).

        Returns
        -------
        ColocatedData
            filtered data object.

        """
        data = self if inplace else self.copy()
        is_2d = data._check_latlon_coords()
        if is_2d:
            data.data = data._filter_country_2d(data.data, region_id, use_country_code)
        else:
            raise NotImplementedError("Cannot yet filter country for 3D ColocatedData object")

        return data

    def apply_latlon_filter(self, lat_range=None, lon_range=None, region_id=None, inplace=False):
        """Apply rectangular latitude/longitude filter

        Parameters
        ----------
        lat_range : list, optional
            latitude range that is supposed to be applied. If specified, then
            also lon_range need to be specified, else, region_id is checked
            against AeroCom default regions (and used if applicable)
        lon_range : list, optional
            longitude range that is supposed to be applied. If specified, then
            also lat_range need to be specified, else, region_id is checked
            against AeroCom default regions (and used if applicable)
        region_id : str
            name of region to be applied. If provided (i.e. not None) then
            input args `lat_range` and `lon_range` are ignored
        inplace : bool, optional
            Apply filter to this object directly or to a copy.
            The default is False.

        Raises
        ------
        ValueError
            if lower latitude bound exceeds upper latitude bound.

        Returns
        -------
        ColocatedData
            filtered data object
        """
        data = self if inplace else self.copy()

        is_2d = data._check_latlon_coords()

        if region_id is not None:
            reg = Region(region_id)
            lon_range = reg.lon_range
            lat_range = reg.lat_range
            region_id = reg.name

        if all([x is None for x in (lat_range, lon_range)]):
            raise ValueError(
                "Please provide input, either for lat_range or lon_range or region_id"
            )
        if lon_range is None:
            lon_range = [-180, 180]
        if lat_range is None:
            lat_range = [-90, 90]

        if lat_range[0] > lat_range[1]:
            raise ValueError(
                f"Lower latitude bound {lat_range[0]} cannot exceed upper "
                f"latitude bound {lat_range[1]}"
            )
        if is_2d:
            filtered = data._filter_latlon_2d(data.data, lat_range, lon_range)
        else:
            filtered = data._filter_latlon_3d(data.data, lat_range, lon_range)

        if not isinstance(region_id, str):
            region_id = "CUSTOM"
        try:
            alt_info = filtered.attrs["filter_name"].split("-", 1)[-1]
        except Exception:
            alt_info = "CUSTOM"

        filtered.attrs["filter_name"] = f"{region_id}-{alt_info}"
        filtered.attrs["region"] = region_id

        data.data = filtered
        return data

    def apply_region_mask(self, region_id, inplace=False):
        """
        Apply a binary regions mask filter to data object. Available binary
        regions IDs can be found at `pyaerocom.const.HTAP_REGIONS`.

        Parameters
        ----------
        region_id : str
            ID of binary regions.
        inplace : bool, optional
            If True, the current instance, is modified, else a new instance
            of `ColocatedData` is created and filtered. The default is False.

        Raises
        ------
        DataCoverageError
            if filtering results in empty data object.

        Returns
        -------
        data : ColocatedData
            Filtered data object.

        """

        data = self if inplace else self.copy()
        arr = data.data
        mask = load_region_mask_xr(region_id)

        if data.ndim == 4:
            mask = mask.interp_like(arr)
            arr = arr.where(mask)
        else:
            # data = data.flatten_latlondim_station_name()

            drop_idx = []
            nstats = len(arr.station_name)
            for (lat, lon, stat) in data._iter_stats():
                if get_mask_value(lat, lon, mask) < 1:
                    drop_idx.append(stat)

            ndrop = len(drop_idx)
            if ndrop == nstats:
                raise DataCoverageError(f"No data available in region {region_id}")
            elif ndrop > 0:
                arr = arr.drop_sel({"station_name": drop_idx})
        data.data = arr
        return data

    def filter_region(self, region_id, check_mask=True, check_country_meta=False, inplace=False):
        """Filter object by region

        Parameters
        ----------
        region_id : str
            ID of region
        inplace : bool
            if True, the filtering is done directly in this instance, else
            a new instance is returned
        check_mask : bool
            if True and region_id a valid name for a binary mask, then the
            filtering is done based on that binary mask.
        check_country_meta : bool
            if True, then the input region_id is first checked against
            available country names in metadata. If that fails, it is assumed
            that this regions is either a valid name for registered rectangular
            regions or for available binary masks.

        Returns
        -------
        ColocatedData
            filtered data object
        """
        filtered = None
        if check_country_meta:
            if region_id in self.countries_available:
                filtered = self.apply_country_filter(
                    region_id, use_country_code=False, inplace=inplace
                )

            elif region_id in self.country_codes_available:
                filtered = self.apply_country_filter(
                    region_id, use_country_code=True, inplace=inplace
                )

        if filtered is None:
            if region_id in const.HTAP_REGIONS:
                filtered = self.apply_region_mask(region_id, inplace)
            elif region_id in REGION_DEFS:
                filtered = self.apply_latlon_filter(region_id=region_id, inplace=inplace)
            else:
                raise UnknownRegion(f"no such region defined {region_id}")
        # if all model and obsdata is NaN raise Exception (this can
        # e.g., be the case if colocate_time=True and all obs is NaN,
        # or if model domain is not covered by the region)
        if np.isnan(filtered.data.data).all():
            # use stats_dummy
            raise DataCoverageError(f"All data is NaN in {region_id}")
        return filtered

    def get_regional_timeseries(self, region_id, **filter_kwargs):
        """
        Compute regional timeseries both for model and obs

        Parameters
        ----------
        region_id : str
            name of region for which regional timeseries is supposed to be
            retrieved
        **filter_kwargs
            additional keyword args passed to :func:`filter_region`.

        Returns
        -------
        dict
            dictionary containing regional timeseries for model (key `mod`)
            and obsdata (key `obs`) and name of region.
        """
        result = {}
        subset = self.filter_region(region_id, inplace=False, **filter_kwargs)
        if subset.has_latlon_dims:
            rgts = subset.data.mean(dim=("latitude", "longitude"))
        else:
            rgts = subset.data.mean(dim="station_name")
        result["obs"] = pd.Series(rgts.data[0], rgts.time.data)
        result["mod"] = pd.Series(rgts.data[1], rgts.time.data)
        result["region"] = region_id
        return result

    def calc_nmb_array(self):
        """Calculate data array with normalised bias (NMB) values

        Returns
        -------
        DataArray
            NMBs at each coordinate
        """
        if not len(self.data_source) == 2:
            raise DataDimensionError("data_source dimension needs exactly 2 entries")
        elif not "time" in self.dims:
            raise DataDimensionError("data needs to have time dimension")
        _arr = self.data
        mod, obs = _arr[1], _arr[0]
        return (mod - obs).sum("time") / obs.sum("time")

    def plot_coordinates(self, marker="x", markersize=12, fontsize_base=10, **kwargs):
        """
        Plot station coordinates

        Uses :func:`pyaerocom.plot.plotcoordinates.plot_coordinates`.

        Parameters
        ----------
        marker : str, optional
            matplotlib marker name used to plot site locations.
            The default is 'x'.
        markersize : int, optional
            Size of site markers. The default is 12.
        fontsize_base : int, optional
            Basic fontsize. The default is 10.
        **kwargs
            additional keyword args passed to
            :func:`pyaerocom.plot.plotcoordinates.plot_coordinates`

        Returns
        -------
        matplotlib.axes.Axes

        """

        from pyaerocom.plot.plotcoordinates import plot_coordinates

        lats, lons = self.get_coords_valid_obs()
        return plot_coordinates(
            lons=lons,
            lats=lats,
            marker=marker,
            markersize=markersize,
            fontsize_base=fontsize_base,
            **kwargs,
        )

    def __contains__(self, val):
        return self.data.__contains__(val)

    def __repr__(self):
        tp = type(self).__name__
        dstr = "<empty>" if self._data is None else repr(self._data)
        return f"pyaerocom.{tp}: data: {dstr}"

    def __str__(self):
        tp = type(self).__name__
        head = f"pyaerocom.{tp}"
        underline = len(head) * "-"
        dstr = "<empty>" if self._data is None else str(self._data)
        return f"{head}\n{underline}\ndata (DataArray): {dstr}"

    ### Deprecated (but still supported) stuff
    # ToDo: v0.12.0
    @property
    def unit(self):
        """DEPRECATED -> use :attr:`units`"""
        warnings.warn(
            "Attr. ColocatedData.unit is deprecated (but still works), "
            "please use ColocatedData.units. "
            "Support guaranteed until pyaerocom v0.12.0",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.units

    @property
    def meta(self):
        """DEPRECATED -> use :attr:`metadata`"""
        warnings.warn(
            "Attr. ColocatedData.meta is deprecated, "
            "use ColocatedData.metadata instread."
            "Support guaranteed until pyaerocom v0.12.0",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.metadata

    @property
    def num_grid_points(self):
        """DEPRECATED -> use :attr:`num_coords`"""
        warnings.warn(
            "ColocatedData.num_grid_points is deprecated, "
            "please use ColocatedData.num_coords"
            "Support guaranteed until pyaerocom v0.12.0",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.num_coords



# isort:skip_file
import logging

logger = logging.getLogger(__name__)


def geopy_available():
    """Helper method that checks if geopy library is available

    Required for import of ReadAeolusL2aData

    Returns
    -------
    bool
        True, if library is available, else False
    """
    try:
        import geopy

        return True
    except ModuleNotFoundError:
        logger.warning("geopy library is not available. Aeolus data read not enabled")
    return False


def coda_available():
    """Helper method that checks if coda library is available

    Required for import of ReadAeolusL2aData and ReadSentinel5pL2Data

    Returns
    -------
    bool
        True, if library is available, else False
    """
    try:
        import coda

        return True
    except ModuleNotFoundError:
        logger.warning(
            "coda library is not available. Sentinel5P and Aeolus data read not enabled"
        )
    return False


from .aerocom_browser import AerocomBrowser
from .ebas_file_index import EbasFileIndex, EbasSQLRequest

# low level EBAS I/O routines
from .ebas_nasa_ames import EbasNasaAmesFile
from .fileconventions import FileConventionRead
from .read_aasetal import ReadAasEtal

# read base classes
from .readgridded import ReadGridded
from .readungriddedbase import ReadUngriddedBase
from .readungridded import ReadUngridded

# Pyaerocom reading interface classes
from .read_aeronet_invv2 import ReadAeronetInvV2
from .read_aeronet_invv3 import ReadAeronetInvV3
from .read_aeronet_sdav2 import ReadAeronetSdaV2
from .read_aeronet_sdav3 import ReadAeronetSdaV3
from .read_aeronet_sunv2 import ReadAeronetSunV2
from .read_aeronet_sunv3 import ReadAeronetSunV3
from .read_airnow import ReadAirNow
from .read_earlinet import ReadEarlinet
from .read_ebas import ReadEbas
from .read_eea_aqerep import ReadEEAAQEREP
from .read_eea_aqerep_v2 import ReadEEAAQEREP_V2
from .read_gaw import ReadGAW
from .read_ghost import ReadGhost
from .read_marcopolo import ReadMarcoPolo
from .read_mscw_ctm import ReadMscwCtm

if geopy_available() and coda_available():
    # the coda and geopy libraries are needed to read l2 data of the supported satellites
    # Aeolus and Sentinel5P
    from .read_aeolus_l2a_data import ReadL2Data
    from .read_sentinel5p_data import ReadL2Data

from . import helpers_units



"""
I/O helper methods of the pyaerocom package
"""
from __future__ import annotations

import logging
import os
import shutil
from datetime import datetime
from importlib import resources
from pathlib import Path
from time import time

import simplejson as json

from pyaerocom import const
from pyaerocom.exceptions import VariableDefinitionError, VarNotAvailableError
from pyaerocom.io import AerocomBrowser

logger = logging.getLogger(__name__)


#: country code file name
#: will be prepended with the path later on
COUNTRY_CODE_FILE = "country_codes.json"


def _check_ebas_db_local_vs_remote(loc_remote, loc_local):
    """
    Check and if applicable, copy ebas_file_index.sqlite3 into cache dir

    Note
    ----
    This may speedup things if remote location is on a mounted server location.
    Nothing the user should worry about in any case.

    Parameters
    ----------
    loc_remote : str
        remote location of ebas_file_index.sqlite3
    loc_local : str
        local (cached) location of ebas_file_index.sqlite3

    Returns
    -------
    str
        valid location of ebas_file_index.sqlite3 that is supposed to be used

    """
    if os.path.exists(loc_remote):  # remote exists
        if os.path.exists(loc_local):
            chtremote = os.path.getmtime(loc_remote)
            chtlocal = os.path.getmtime(loc_local)
            if chtlocal == chtremote:
                return loc_local

        # changing time differs -> try to copy to local and if that
        # fails, use remote location
        try:
            t0 = time()
            shutil.copy2(loc_remote, loc_local)
            logger.info(
                f"Copied EBAS SQL database to {loc_local}\nElapsed time: {time()-t0:.3f} s"
            )

            return loc_local
        except Exception as e:
            logger.warning(f"Failed to copy EBAS SQL database. Reason: {repr(e)}")
            return loc_remote
    return loc_remote


def aerocom_savename(data_id, var_name, vert_code, year, ts_type):
    """Generate filename in AeroCom conventions

    ToDo: complete docstring
    """
    return f"aerocom3_{data_id}_{var_name}_{vert_code}_{year}_{ts_type}.nc"


def _print_read_info(i, mod, tot_num, last_t, name, logger):
    """Helper for displaying standardised output in reading classes

    Not to be used directly
    """
    t = datetime.now()
    logger.info(
        f"Reading files {i+1}-{i+1+mod} of {tot_num} "
        f"({name}) | {t:%T} (delta = {(t-last_t).seconds} s')"
    )
    return t


def get_metadata_from_filename(filename):
    """Try access metadata information from filename"""
    from pyaerocom.io.fileconventions import FileConventionRead

    fc = FileConventionRead().from_file(filename)
    return fc.get_info_from_file(filename)


def read_ebas_flags_file(ebas_flags_csv):
    """Reads file ebas_flags.csv

    Parameters
    ----------
    ebas_flags_csv : str
        file containing flag info

    Returns
    -------
    dict
        dict with loaded flag info
    """
    valid = {}
    values = {}
    info = {}
    with open(ebas_flags_csv) as fio:
        for line in fio:
            spl = line.strip().split(",")
            num = int(spl[0].strip())
            try:
                val_str = spl[-1][1:-1]
            except Exception:
                raise OSError(
                    f"Failed to read flag information in row {line} "
                    f"(Check if entries in ebas_flags.csv are quoted)"
                )
            info_str = ",".join(spl[1:-1])
            try:
                info_str = info_str[1:-1]
            except Exception:
                raise OSError(
                    f"Failed to read flag information in row {line} "
                    f"(Check if entries in ebas_flags.csv are quoted)"
                )
            isvalid = True if val_str == "V" else False
            valid[num] = isvalid
            values[num] = val_str
            info[num] = info_str
    result = {}
    result["valid"] = valid
    result["info"] = info
    result["vals"] = values
    return result


def add_file_to_log(filepath, err_msg):
    """
    Add input file path to error logdir

    The logdir location can be accessed via :attr:`pyaerocom.const.LOGFILESDIR`

    Parameters
    ----------
    filepath : str or Path
        path of file that has an error
    err_msg : str
        Problem associated with input file

    """
    if isinstance(filepath, Path):
        filepath = str(filepath)
    try:
        dirname = os.path.dirname(filepath)
        spl = dirname.split(os.sep)
        if spl[-1].lower() == "renamed":
            model_or_obs_id = spl[-2]
        else:
            model_or_obs_id = spl[-1]
    except Exception:
        model_or_obs_id = "others"

    logdir = const.LOGFILESDIR

    logfile = os.path.join(logdir, f"{model_or_obs_id}.log")

    if os.path.exists(logfile):  # check if this file is already flagged
        with open(logfile) as f:
            for line in f:
                if filepath == line.strip():
                    return  # file is already flagged -> ignore

    logfile_err = os.path.join(logdir, f"{model_or_obs_id}_ERR.log")
    with open(logfile, "a+") as f:
        f.write(f"{filepath}\n")
    with open(logfile_err, "a+") as ferr:
        ferr.write(f"{filepath}\n{err_msg}\n\n")


def get_standard_name(var_name):
    """Get standard name of aerocom variable

    Parameters
    ----------
    var_name : str
        HTAP2 variable name

    Returns
    --------
    str
        corresponding standard name

    Raises
    ------
    VarNotAvailableError
        if input variable is not defined in *variables.ini* file
    VariableDefinitionError
        if standarad name is not set for variable in *variables.ini* file
    """
    if not var_name in const.VARS:
        raise VarNotAvailableError(f"No such variable {var_name}. Check variables.ini")
    name = const.VARS[var_name].standard_name
    if name is None:
        raise VariableDefinitionError("standard_name not defined for variable")
    return name


def search_data_dir_aerocom(name_or_pattern, ignorecase=True):
    """Search Aerocom data directory based on model / data ID"""
    browser = AerocomBrowser()
    return browser.find_data_dir(name_or_pattern, ignorecase)


def get_all_supported_ids_ungridded():
    """Get list of datasets that are supported by :class:`ReadUngridded`

    Returns
    -------
    list
        list with supported network names
    """
    from pyaerocom.io import ReadUngridded

    return ReadUngridded().SUPPORTED_DATASETS


def get_obsnetwork_dir(obs_id):
    """Returns data path for obsnetwork ID

    Parameters
    ----------
    obs_id : str
        ID  of obsnetwork (e.g. AeronetSunV2Lev2.daily)

    Returns
    -------
    str
        corresponding directory from ``pyaerocom.const``

    Raises
    ------
    ValueError
        if obs_id is invalid
    IOError
        if directory does not exist
    """
    if not obs_id in const.OBSLOCS_UNGRIDDED:
        raise ValueError(f"Observation network ID {obs_id} does not exist")

    data_dir = const.OBSLOCS_UNGRIDDED[obs_id]
    if not os.path.exists(data_dir):
        raise FileNotFoundError(
            f"Data directory {data_dir} for observation network {obs_id} does not exist"
        )
    return data_dir


def get_country_name_from_iso(
    iso_code: str | None = None, filename: str | Path | None = None, return_as_dict: bool = False
):
    """get the country name from the 2 digit iso country code

    the underlaying json file was taken from this github repository
    https://github.com/lukes/ISO-3166-Countries-with-Regional-Codes

    Parameters
    ----------
    iso_code : :obj:`str`
        string containing the 2 character iso code of the country (e.g. no for Norway)
    filename : :obj:`str` , optional
        optional string with the json file to read
    return_as_dict : :obj:`bool`, optional
        flag to get the entire list of countries as a dictionary with the country codes
        as keys and the country names as value
        Useful if you have to get the names for a lot of country codes

    Returns
    -------
    string with country name or dictionary with iso codes as keys and the country names as values
    empty string if the country code was not found


    Raises
    ------
    ValueError
        if the country code ins invalid
    """
    if filename is None:
        # set default file name
        with resources.path("pyaerocom.data", COUNTRY_CODE_FILE) as path:
            filename = path

    if isinstance(filename, str):
        filename = Path(filename)
    json_data = json.loads(filename.read_text())

    iso_dict = {}
    for indict in json_data:
        iso_dict[indict["alpha-2"]] = indict["name"]

    if iso_code is None or return_as_dict:
        return iso_dict

    return iso_dict[iso_code.upper()]



import glob
import logging
import os
import re

import numpy as np
import xarray as xr

from pyaerocom import const
from pyaerocom.exceptions import VarNotAvailableError
from pyaerocom.griddeddata import GriddedData
from pyaerocom.io._read_mscw_ctm_helpers import (
    add_dataarrays,
    calc_concNhno3,
    calc_concNnh3,
    calc_concNnh4,
    calc_concNno3pm10,
    calc_concNno3pm25,
    calc_concNtnh,
    calc_concsspm25,
    calc_conNtno3,
    calc_vmrno2,
    calc_vmrox,
    subtract_dataarrays,
    update_EC_units,
)
from pyaerocom.units_helpers import UALIASES
from pyaerocom.variable_helpers import get_emep_variables

logger = logging.getLogger(__name__)


class ReadMscwCtm:
    """
    Class for reading model output from the EMEP MSC-W chemical transport model.

    Parameters
    ----------
    filepath : str
        Path to netcdf file.
    data_id : str
        string ID of model (e.g. "AATSR_SU_v4.3","CAM5.3-Oslo_CTRL2016")
    data_dir : str, optional
        Base directory of EMEP data, containing one or more netcdf files

    Attributes
    ----------
    data_id : str
        ID of model
    filename : str
        name of data file to be read.
    """

    # dictionary containing information about additionally required variables
    # for each auxiliary variable (i.e. each variable that is not provided
    # by the original data but computed on import)
    AUX_REQUIRES = {
        "depso4": ["dryso4", "wetso4"],
        "concbc": ["concbcf", "concbcc"],
        "concno3": ["concno3c", "concno3f"],
        "concoa": ["concoac", "concoaf"],
        "concpmgt25": ["concpm10", "concpm25"],
        "concNhno3": ["conchno3"],
        "concNnh3": ["concnh3"],
        "concNnh4": ["concnh4"],
        "concNno3pm10": ["concno3f", "concno3c"],
        "concNno3pm25": ["concno3f", "concno3c"],
        "concNtno3": ["conchno3", "concno3f", "concno3c"],
        "concNtnh": ["concnh3", "concnh4"],
        "concsspm25": ["concssf", "concssc"],
        "concsspm10": ["concssf", "concssc"],
        "concCecpm25": ["concecpm25"],
        "vmrox": ["concno2", "vmro3"],
        "vmrno2": ["concno2"],
    }

    # Functions that are used to compute additional variables (i.e. one
    # for each variable defined in AUX_REQUIRES)
    # NOTE: these methods are supposed to work for xarray.DataArray instances
    # not iris.cube.Cube instance
    AUX_FUNS = {
        "depso4": add_dataarrays,
        "concbc": add_dataarrays,
        "concno3": add_dataarrays,
        "concoa": add_dataarrays,
        "concpmgt25": subtract_dataarrays,
        "concNhno3": calc_concNhno3,
        "concNnh3": calc_concNnh3,
        "concNnh4": calc_concNnh4,
        "concNno3pm10": calc_concNno3pm10,
        "concNno3pm25": calc_concNno3pm25,
        "concNtno3": calc_conNtno3,
        "concNtnh": calc_concNtnh,
        "concsspm25": calc_concsspm25,
        "concsspm10": add_dataarrays,
        "concCecpm25": update_EC_units,
        "vmrox": calc_vmrox,
        "vmrno2": calc_vmrno2,
    }

    #: supported filename masks, placeholder is for frequencies
    FILE_MASKS = ["Base_*.nc"]

    #: frequencies encoded in filenames
    FREQ_CODES = {
        "hour": "hourly",
        "day": "daily",
        "month": "monthly",
        "fullrun": "yearly",
    }

    REVERSE_FREQ_CODES = {
        "hourly": "hour",
        "daily": "day",
        "monthly": "month",
        "yearly": "fullrun",
    }

    DEFAULT_FILE_NAME = "Base_day.nc"

    def __init__(self, data_id=None, data_dir=None):
        self._data_dir = None
        # opened dataset (for performance boost), will be reset if data_dir is
        # changed
        self._filename = None
        self._filedata = None
        self._filepaths = None
        self._filepath = None

        self._file_mask = self.FILE_MASKS[0]
        self._files = None

        self.var_map = get_emep_variables()

        if data_dir is not None:
            if not isinstance(data_dir, str) or not os.path.exists(data_dir):
                raise FileNotFoundError(f"{data_dir}")

            self.data_dir = data_dir

        self.data_id = data_id

        self.filename = self.DEFAULT_FILE_NAME

    def search_all_files(self):
        namelist = self._get_namelist_from_folder()
        self.filepaths = self._get_files_from_namelist(namelist)

    def _get_files_from_namelist(self, namelist):
        files = []
        for d in namelist:
            mask, f = self._check_files_in_data_dir(d)
            files += f

        return files

    def _get_namelist_from_folder(self):
        """
        Finds all the subfolders where a emep file for one year might be.

        Note
        -------
        Checks only current level for folders. Should be able
        to search deeper.

        The only qualification of being a valid subfolder is whether or not
        the subfolder has contains a number >= 2000. There are no check if there
        are any emep files in the folder

        Returns
        -------
        List
         List of the names of the subfolder

        """
        dd = self.data_dir

        dirs = glob.glob(dd + "/*/")
        namelist = []
        yrs = []

        for d in dirs:
            if re.match(r".*20\d\d.*", d) is None:
                continue
            yrs.append(d.split("/")[-1])

            namelist.append(d)

        if len(namelist) == 0:
            namelist = [dd]
        else:
            namelist = [d for _, d in sorted(zip(yrs, namelist))]
        return list(set(namelist))

    def _get_yrs_from_filepaths(self):
        fps = self.filepaths
        yrs = []
        for fp in fps:

            try:
                yr = re.search(r".*(20\d\d).*", fp).group(1)
            except:
                raise ValueError(f"Could not find any year in {fp}")

            yrs.append(yr)

        return sorted(list(set(yrs)))

    def _get_tst_from_file(self, file):
        # tst = re.search("Base_(.*).nc", file).group(1)
        mask = self._file_mask.replace("*", "(.*)")
        tst = re.search(mask, file).group(1)

        if "LF_" in tst:
            return None

        if tst not in list(self.FREQ_CODES):
            raise ValueError(f"The ts_type {tst} is not supported")

        return self.FREQ_CODES[tst]

    def _clean_filepaths(self, filepaths, yrs, ts_type):
        clean_paths = []
        found_yrs = []

        yrs = [int(yr) for yr in yrs]
        for path in filepaths:
            ddir, file = os.path.split(path)

            if self._get_tst_from_file(file) != ts_type:
                continue

            yrs_dir = ddir.split(os.sep)[-1]
            try:
                yr = re.search(r".*(20\d\d).*", yrs_dir).group(1)
            except:
                raise ValueError(f"Could not find any year in {yrs_dir}")

            if int(yr) not in yrs:
                raise ValueError(f"The year {yr} is not in {yrs}")

            if int(yr) in found_yrs:
                raise ValueError(f"The year {yr} is already found: {found_yrs}")

            found_yrs.append(int(yr))

            clean_paths.append(path)

        if len(found_yrs) != len(yrs):
            raise ValueError(
                f"A different amount of years {found_yrs} were found compared tp {yrs}"
            )

        return [d for _, d in sorted(zip(found_yrs, clean_paths))]

    @property
    def data_dir(self):
        """
        Directory containing netcdf files
        """
        if self._data_dir is None:
            raise AttributeError(f"data_dir needs to be set before accessing")
        return self._data_dir

    @data_dir.setter
    def data_dir(self, val):
        if val is None:
            raise ValueError(f"Data dir {val} needs to be a dictionary or a file")
        if not os.path.isdir(val):
            raise FileNotFoundError(val)
        self._file_mask = self.FILE_MASKS[0]
        self._data_dir = val
        self._filedata = None
        self.search_all_files()
        self._files = self.filepaths

    @property
    def filename(self):
        """
        Name of netcdf file
        """
        return self._filename

    @filename.setter
    def filename(self, val):
        """
        Name of netcdf file
        """
        if not isinstance(val, str):
            raise ValueError("needs str")
        elif val == self._filename:
            return
        self._filename = val
        self._filedata = None

    @property
    def filepath(self):
        """
        Path to data file
        """
        if self.data_dir is None and self._filepaths is None:
            raise AttributeError("data_dir or filepaths needs to be set before accessing")
        return self._filepath

    @filepath.setter
    def filepath(self, value):
        if not isinstance(value, str):
            raise TypeError("needs to be a string")

        self._filepath = value
        ddir, fname = os.path.split(value)
        self.data_dir = ddir
        self.filename = fname

    @property
    def filepaths(self):
        """
        Path to data file
        """
        if self.data_dir is None and self._filepaths is None:
            raise AttributeError("data_dir or filepaths needs to be set before accessing")
        return self._filepaths

    @filepaths.setter
    def filepaths(self, value):
        if not isinstance(value, list):
            raise ValueError("needs to be list of strings")
        self._filepaths = value

    @property
    def filedata(self):
        """
        Loaded netcdf file (:class:`xarray.Dataset`)
        """
        if self._filedata is None:
            self.open_file()
        return self._filedata

    def _check_files_in_data_dir(self, data_dir):
        """
        Check for data files in input data directory

        Parameters
        ----------
        data_dir : str
            directory to be searched.

        Raises
        ------
        FileNotFoundError
            if no EMEP files can be identified

        Returns
        -------
        str
            file mask used (is identified automatically based on
            :attr:`FILE_MASKS`)
        list
            list of file matches

        """

        for fmask in self.FILE_MASKS:
            matches = glob.glob(f"{data_dir}/{fmask}")
            if len(matches) > 0:
                return fmask, matches
        raise FileNotFoundError(
            f"No valid model files could be found in {data_dir} for any of the "
            f"supported file masks: {self.FILE_MASKS}"
        )

    @property
    def ts_type(self):
        """
        Frequency of time dimension of current data file

        Raises
        ------
        AttributeError
            if :attr:`filename` is not set.

        Returns
        -------
        str
            current ts_type.

        """
        return self.ts_type_from_filename(self.filename)

    @property
    def ts_types(self):
        """
        List of available frequencies

        Raises
        ------
        AttributeError
            if :attr:`data_dir` is not set.

        Returns
        -------
        list
            list of available frequencies

        """
        if not isinstance(self._files, list):
            raise AttributeError("please set data_dir first")
        tsts = []
        for file in self._files:
            tsts.append(self.ts_type_from_filename(file))
        return list(set(tsts))

    @property
    def years_avail(self):
        """
        Years available in loaded dataset
        """
        data = self.filepaths
        years = self._get_yrs_from_filepaths()

        years = list(np.unique(years))
        return sorted(years)

    @property
    def vars_provided(self):
        """Variables provided by this dataset"""
        return list(self.var_map) + list(self.AUX_REQUIRES)

    def open_file(self):
        """
        Open current netcdf file

        Returns
        -------
        dict(xarray.Dataset)
            Dict with years as keys and Datasets as items

        """
        fps = self.filepaths
        ds = {}

        yrs = self._get_yrs_from_filepaths()

        ts_type = self.ts_type_from_filename(self.filename)
        fps = self._clean_filepaths(fps, yrs, ts_type)

        for i, fp in enumerate(fps):
            if not os.path.split(fp)[-1] == self.filename:
                continue

            logger.info(f"Opening {fp}")
            tmp_ds = xr.open_dataset(fp)

            ds[yrs[i]] = tmp_ds

        self._filedata = ds

        return ds

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "ReadMscwCtm"

    def has_var(self, var_name):
        """Check if variable is supported

        Parameters
        ----------
        var_name : str
            variable to be checked

        Returns
        -------
        bool
        """
        avail = self.vars_provided
        if var_name in avail or const.VARS[var_name].var_name_aerocom in avail:
            return True
        return False

    def ts_type_from_filename(self, filename):
        """
        Get ts_type from filename

        Parameters
        ----------
        filename : str

        Raises
        ------
        ValueError
            if ts_type cannot be inferred from filename.

        Returns
        -------
        tstype : str
        """
        filename = os.path.basename(filename)
        for substr, tstype in self.FREQ_CODES.items():
            if substr in filename:
                return tstype
        raise ValueError(f"Failed to retrieve ts_type from filename {filename}")

    def filename_from_ts_type(self, ts_type):
        """
        Infer file name of data based on input ts_type

        Parameters
        ----------
        ts_type : str
            desired time freq of data

        Raises
        ------
        ValueError
            If no file could be inferred.

        Returns
        -------
        fname : str
            Name of data file inferred based on current file mask and input
            freq.

        """
        mask = self._file_mask
        for substr, tst in self.FREQ_CODES.items():
            if tst == ts_type:
                fname = mask.replace("*", substr)
                return fname
        raise ValueError(f"failed to infer filename from input ts_type={ts_type}")

    def _compute_var(self, var_name_aerocom, ts_type):
        """Compute auxiliary variable

        Like :func:`read_var` but for auxiliary variables
        (cf. AUX_REQUIRES)

        Parameters
        ----------
        var_name : str
            variable that are supposed to be read
        ts_type : str
            string specifying temporal resolution.

        Returns
        -------
        GriddedData
            loaded data object
        """

        temp_arrs = []
        req = self.AUX_REQUIRES[var_name_aerocom]
        aux_func = self.AUX_FUNS[var_name_aerocom]
        logger.info(f"computing {var_name_aerocom} from {req} using {aux_func}")
        for aux_var in self.AUX_REQUIRES[var_name_aerocom]:
            arr = self._load_var(aux_var, ts_type)
            temp_arrs.append(arr)

        return aux_func(*temp_arrs)

    def _load_var(self, var_name_aerocom, ts_type):
        """
        Load variable data as :class:`xarray.DataArray`.

        This combines both, variables that can be read directly and auxiliary
        variables that are computed.

        Parameters
        ----------
        var_name_aerocom : str
            variable name
        ts_type : str
            desired frequency

        Raises
        ------
        VarNotAvailableError
            if input variable is not available

        Returns
        -------
        xarray.DataArray
            loaded data

        """
        if var_name_aerocom in self.var_map:  # can be read
            return self._read_var_from_file(var_name_aerocom, ts_type)
        elif var_name_aerocom in self.AUX_REQUIRES:
            return self._compute_var(var_name_aerocom, ts_type)
        raise VarNotAvailableError(f"Variable {var_name_aerocom} is not supported")

    def read_var(self, var_name, ts_type=None, **kwargs):
        """Load data for given variable.

        Parameters
        ----------
        var_name : str
            Variable to be read
        ts_type : str
            Temporal resolution of data to read. Supported are
            "hourly", "daily", "monthly" , "yearly".

        Returns
        -------
        GriddedData
        """
        if not self.has_var(var_name):
            raise VarNotAvailableError(var_name)
        var = const.VARS[var_name]
        var_name_aerocom = var.var_name_aerocom

        if self.data_dir is None:
            raise ValueError("data_dir must be set before reading.")
        elif self.filename is None and ts_type is None:
            raise ValueError("please specify ts_type")
        elif ts_type is not None:
            # filename and ts_type are set. update filename if ts_type suggests
            # that current file has different resolution
            self.filename = self.filename_from_ts_type(ts_type)

        ts_type = self.ts_type

        arr = self._load_var(var_name_aerocom, ts_type)
        if arr.units in UALIASES:
            arr.attrs["units"] = UALIASES[arr.units]
        try:
            cube = arr.to_iris()
        except MemoryError as e:
            raise NotImplementedError(f"BAAAM: {e}")

        if ts_type == "hourly":
            cube.coord("time").convert_units("hours since 1900-01-01")
        gridded = GriddedData(
            cube,
            var_name=var_name_aerocom,
            ts_type=ts_type,
            check_unit=True,
            convert_unit_on_init=True,
        )

        #!obsolete
        # if var.is_deposition:
        #    implicit_to_explicit_rates(gridded, ts_type)

        # At this point a GriddedData object with name gridded should exist

        gridded.metadata["data_id"] = self.data_id
        gridded.metadata["from_files"] = self.filepaths

        # Remove unneccessary metadata. Better way to do this?
        for metadata in ["current_date_first", "current_date_last"]:
            if metadata in gridded.metadata.keys():
                del gridded.metadata[metadata]
        return gridded

    def _read_var_from_file(self, var_name_aerocom, ts_type):
        """
        Read variable data from file as :class:`xarray.DataArray`.

        See also :func:`_load_var`

        Parameters
        ----------
        var_name_aerocom : str
            variable name
        ts_type : str
            desired frequency

        Raises
        ------
        VarNotAvailableError
            if input variable is not available

        Returns
        -------
        xarray.DataArray
            loaded data

        """
        emep_var = self.var_map[var_name_aerocom]

        try:
            filedata = self.filedata

            if len(filedata) == 1:
                data = filedata[list(filedata)[0]][emep_var]
            else:
                if ts_type == "hourly":
                    raise ValueError(
                        f"ts_type {ts_type} can not be hourly when using multiple years"
                    )
                data = xr.concat([ds[emep_var] for ds in filedata.values()], dim="time")

        except KeyError:
            raise VarNotAvailableError(
                f"{var_name_aerocom} ({emep_var}) not available in {self.filename}"
            )
        data.attrs["long_name"] = var_name_aerocom
        data.time.attrs["long_name"] = "time"
        data.time.attrs["standard_name"] = "time"
        prefix = emep_var.split("_")[0]
        data.attrs["units"] = self.preprocess_units(data.units, prefix)
        return data

    @staticmethod
    def preprocess_units(units, prefix):
        """
        Update units for certain variables

        Parameters
        ----------
        units : str
            Current unit of data
        prefix : str, optional
            Variable prefix (e.g. AOD, AbsCoeff).

        Returns
        -------
        str
            updated unit (where applicable)

        """
        if units == "" and prefix == "AOD":  #
            return "1"
        elif units == "" and prefix == "AbsCoef":
            return "m-1"
        return units


class ReadEMEP(ReadMscwCtm):
    """Old name of :class:`ReadMscwCtm`."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        print("You are using a deprecated name ReadEMEP for class ReadMscwCtm")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
from multiprocessing import dummy

from pyaerocom.aeroval._processing_base import HasColocator, ProcessingEngine
from pyaerocom.aeroval.coldatatojson_engine import ColdataToJsonEngine
from pyaerocom.aeroval.helpers import delete_dummy_model, make_dummy_model
from pyaerocom.aeroval.modelmaps_engine import ModelMapsEngine
from pyaerocom.aeroval.superobs_engine import SuperObsEngine

logger = logging.getLogger(__name__)


class ExperimentProcessor(ProcessingEngine, HasColocator):
    """Processing engine for AeroVal experiment

    By default, this class processes one configuration file, represented by
    :class:`EvalSetup`. As such, an instance of :class:`EvalSetup` represents
    an AeroVal experiment, comprising a list of models, a list of observations
    (and variables).

    For each possible (or defined) model / obs / variable combination, the
    processing engine will perform spatial and temporal co-location and will
    store on co-located NetCDF file (e.g. if there are 2 models, 2 observation
    networks and 2 variables there will be 4 co-located NetCDF files).
    The co-location is done using :class:`pyaerocom.colocation_auto.Colocator`.

    """

    def _run_single_entry(self, model_name, obs_name, var_list):
        if model_name == obs_name:
            msg = f"Cannot run same dataset against each other ({model_name} vs. {obs_name})"
            logger.info(msg)
            return
        ocfg = self.cfg.get_obs_entry(obs_name)
        if ocfg["is_superobs"]:
            try:
                engine = SuperObsEngine(self.cfg)
                engine.run(
                    model_name=model_name,
                    obs_name=obs_name,
                    var_list=var_list,
                    try_colocate_if_missing=True,
                )
            except Exception:
                if self.raise_exceptions:
                    raise
                logger.warning("failed to process superobs...")
        elif ocfg["only_superobs"]:
            logger.info(
                f"Skipping json processing of {obs_name}, as this is "
                f"marked to be used only as part of a superobs "
                f"network"
            )
        else:
            col = self.get_colocator(model_name, obs_name)
            if self.cfg.processing_opts.only_json:
                files_to_convert = col.get_available_coldata_files(var_list)
            else:
                col.run(var_list)
                files_to_convert = col.files_written

            if self.cfg.processing_opts.only_colocation:
                logger.info(
                    f"FLAG ACTIVE: only_colocation: Skipping "
                    f"computation of json files for {obs_name} /"
                    f"{model_name} combination."
                )
            else:
                engine = ColdataToJsonEngine(self.cfg)
                engine.run(files_to_convert)

    def run(self, model_name=None, obs_name=None, var_list=None, update_interface=True):
        """Create colocated data and json files for model / obs combination

        Parameters
        ----------
        model_name : str or list, optional
            Name or pattern specifying model that is supposed to be analysed.
            Can also be a list of names or patterns to specify multiple models.
            If None (default), then all models are run that are part of this
            experiment.
        obs_name : :obj:`str`, or :obj:`list`, optional
            Like :attr:`model_name`, but for specification(s) of observations
            that are supposed to be used. If None (default) all observations
            are used.
        var_list : list, optional
            list variables supposed to be analysed. If None, then all
            variables available are used. Defaults to None. Can also be
            `str` type.
        update_interface : bool
            if true, relevant json files that determine what is displayed
            online are updated after the run, including the the menu.json file
            and also, the model info table (minfo.json) file is created and
            saved in :attr:`exp_dir`.

        Returns
        -------
        list
            list containing all colocated data objects that have been converted
            to json files.
        """
        if isinstance(var_list, str):
            var_list = [var_list]

        self.cfg._check_time_config()

        obs_list = self.cfg.obs_cfg.keylist(obs_name)
        if not self.cfg.model_cfg:
            logging.info("No model found, will make dummy model data")
            self.cfg.webdisp_opts.hide_charts = ["scatterplot"]
            self.cfg.webdisp_opts.hide_pages = ["maps.php", "intercomp.php", "overall.php"]
            model_id = make_dummy_model(obs_list, self.cfg)
            self.cfg.processing_opts.obs_only = True
            use_dummy_model = True
        else:
            model_id = None
            use_dummy_model = False

        model_list = self.cfg.model_cfg.keylist(model_name)

        logger.info("Start processing")

        # compute model maps (completely independent of obs-eval
        # processing below)
        if self.cfg.webdisp_opts.add_model_maps:
            engine = ModelMapsEngine(self.cfg)
            engine.run(model_list=model_list, var_list=var_list)

        if not self.cfg.processing_opts.only_model_maps:
            for obs_name in obs_list:
                for model_name in model_list:
                    self._run_single_entry(model_name, obs_name, var_list)

        if update_interface:
            self.update_interface()
        if use_dummy_model:
            delete_dummy_model(model_id)
        logger.info("Finished processing.")

    def update_interface(self):
        """Update aeroval interface

        Things done here:

            - Update menu file
            - Make aeroval info table json (tab informations in interface)
            - update and order heatmap file
        """
        self.exp_output.update_interface()



"""
Mathematical low level utility methods of pyaerocom
"""
import numpy as np
from scipy.stats import kendalltau, pearsonr, spearmanr

from pyaerocom._warnings import ignore_warnings

### LAMBDA FUNCTIONS
in_range = lambda x, low, high: low <= x <= high

### OTHER FUNCTIONS


def is_strictly_monotonic(iter1d) -> bool:
    """
    Check if 1D iterble is strictly monotonic

    Parameters
    ----------
    iter1d
        1D iterable object to be tested

    Returns
    -------
    bool

    """
    return True if np.all(np.diff(iter1d) > 0) else False


def make_binlist(vmin: float, vmax: float, num: int = None) -> list:
    """"""
    if num is None:
        num = 8
    return list(np.linspace(vmin, vmax, num + 1))


def weighted_sum(data, weights):
    """Compute weighted sum using numpy dot product

    Parameters
    ----------
    data : ndarray
        data array that is supposed to be summed up
    weights : ndarray
        array containing weights for each point in `data`

    Returns
    -------
    float
        weighted sum of values in input array
    """
    return np.dot(data, weights)


def sum(data, weights=None):
    """Summing operation with option to perform weighted sum

    Parameters
    ----------
    data : ndarray
        data array that is supposed to be summed up
    weights : ndarray, optional
        array containing weights for each point in `data`

    Returns
    -------
    float or int
        sum of values in input array
    """
    if weights is None:
        return np.sum(data)
    return weighted_sum(data, weights)


def weighted_mean(data, weights):
    """Compute weighted mean

    Parameters
    ----------
    data : ndarray
        data array that is supposed to be averaged
    weights : ndarray
        array containing weights for each point in `data`

    Returns
    -------
    float or int
        weighted mean of data array
    """
    return np.sum(data * weights) / np.sum(weights)


def weighted_cov(ref_data, data, weights):
    """Compute weighted covariance

    Parameters
    ----------
    data_ref : ndarray
        x data
    data : ndarray
        y data
    weights : ndarray
        array containing weights for each point in `data`

    Returns
    -------
    float
        covariance
    """
    avgx = weighted_mean(ref_data, weights)
    avgy = weighted_mean(data, weights)
    return np.sum(weights * (ref_data - avgx) * (data - avgy)) / np.sum(weights)


def weighted_corr(ref_data, data, weights):
    """Compute weighted correlation

    Parameters
    ----------
    data_ref : ndarray
        x data
    data : ndarray
        y data
    weights : ndarray
        array containing weights for each point in `data`

    Returns
    -------
    float
       weighted correlation coefficient
    """
    wcovxy = weighted_cov(ref_data, data, weights)

    wcovxx = weighted_cov(ref_data, ref_data, weights)
    wcovyy = weighted_cov(data, data, weights)
    wsigmaxy = np.sqrt(wcovxx * wcovyy)
    return wcovxy / wsigmaxy


@ignore_warnings(
    RuntimeWarning, "invalid value encountered in double_scalars", "An input array is constant"
)
def corr(ref_data, data, weights=None):
    """Compute correlation coefficient

    Parameters
    ----------
    data_ref : ndarray
        x data
    data : ndarray
        y data
    weights : ndarray, optional
        array containing weights for each point in `data`

    Returns
    -------
    float
       correlation coefficient
    """
    if weights is None:
        return pearsonr(ref_data, data)[0]
    return weighted_corr(ref_data, data, weights)


def _nanmean_and_std(data):
    """
    Calculate mean and std for input data (may contain NaN's')

    Parameters
    ----------
    data : list or numpy.ndarray
        input data

    Returns
    -------
    float
        mean value of input data.
    float
        standard deviation of input data.

    """
    if np.all(np.isnan(data)):
        return (np.nan, np.nan)
    return (np.nanmean(data), np.nanstd(data))


@ignore_warnings(
    RuntimeWarning, "An input array is constant", "invalid value encountered in true_divide"
)
def calc_statistics(data, ref_data, lowlim=None, highlim=None, min_num_valid=1, weights=None):
    """Calc statistical properties from two data arrays

    Calculates the following statistical properties based on the two provided
    1-dimensional data arrays and returns them in a dictionary (keys are
    provided after the arrows):

        - Mean value of both arrays -> refdata_mean, data_mean
        - Standard deviation of both arrays -> refdata_std, data_std
        - RMS (Root mean square) -> rms
        - NMB (Normalised mean bias) -> nmb
        - MNMB (Modified normalised mean bias) -> mnmb
        - FGE (Fractional gross error) -> fge
        - R (Pearson correlation coefficient) -> R
        - R_spearman (Spearman corr. coeff) -> R_spearman

    Note
    ----
    Nans are removed from the input arrays, information about no. of removed
    points can be inferred from keys `totnum` and `num_valid` in return dict.

    Parameters
    ----------
    data : ndarray
        array containing data, that is supposed to be compared with reference
        data
    ref_data : ndarray
        array containing data, that is used to compare `data` array with
    lowlim : float
        lower end of considered value range (e.g. if set 0, then all datapoints
        where either ``data`` or ``ref_data`` is smaller than 0 are removed)
    highlim : float
        upper end of considered value range
    min_num_valid : int
        minimum number of valid measurements required to compute statistical
        parameters.

    Returns
    -------
    dict
        dictionary containing computed statistics

    Raises
    ------
    ValueError
        if either of the input arrays has dimension other than 1

    """
    data = np.asarray(data)
    ref_data = np.asarray(ref_data)

    if not data.ndim == 1 or not ref_data.ndim == 1:
        raise ValueError("Invalid input. Data arrays must be one dimensional")

    result = {}

    mask = ~np.isnan(ref_data) * ~np.isnan(data)
    num_points = mask.sum()

    data, ref_data = data[mask], ref_data[mask]

    weighted = False if weights is None else True

    result["totnum"] = float(len(mask))
    result["num_valid"] = float(num_points)
    ref_mean, ref_std = _nanmean_and_std(ref_data)
    data_mean, data_std = _nanmean_and_std(data)
    result["refdata_mean"] = ref_mean
    result["refdata_std"] = ref_std
    result["data_mean"] = data_mean
    result["data_std"] = data_std
    result["weighted"] = weighted

    if not num_points >= min_num_valid:
        if lowlim is not None:
            valid = np.logical_and(data > lowlim, ref_data > lowlim)
            data = data[valid]
            ref_data = ref_data[valid]
        if highlim is not None:
            valid = np.logical_and(data < highlim, ref_data < highlim)
            data = data[valid]
            ref_data = ref_data[valid]

        result["rms"] = np.nan
        result["nmb"] = np.nan
        result["mnmb"] = np.nan
        result["fge"] = np.nan
        result["R"] = np.nan
        result["R_spearman"] = np.nan

        return result

    if lowlim is not None:
        valid = np.logical_and(data > lowlim, ref_data > lowlim)
        data = data[valid]
        ref_data = ref_data[valid]
    if highlim is not None:
        valid = np.logical_and(data < highlim, ref_data < highlim)
        data = data[valid]
        ref_data = ref_data[valid]

    difference = data - ref_data

    diffsquare = difference**2

    if weights is not None:
        weights = weights[mask]
        weights = weights / weights.max()
        result[
            "NOTE"
        ] = "Weights were not applied to FGE and kendall and spearman corr (not implemented)"

    result["rms"] = np.sqrt(np.average(diffsquare, weights=weights))

    # NO implementation to apply weights yet ...
    if num_points > 1:
        result["R"] = corr(data, ref_data, weights)
        result["R_spearman"] = spearmanr(data, ref_data)[0]
        result["R_kendall"] = kendalltau(data, ref_data)[0]
    else:
        result["R"] = np.nan
        result["R_spearman"] = np.nan
        result["R_kendall"] = np.nan

    sum_diff = sum(difference, weights=weights)
    sum_refdata = sum(ref_data, weights=weights)

    if sum_refdata == 0:
        if sum_diff == 0:
            nmb = 0
        else:
            nmb = np.nan
    else:
        nmb = sum_diff / sum_refdata

    sum_data_refdata = data + ref_data
    # for MNMB, and FGE: don't divide by 0 ...
    mask = ~np.isnan(sum_data_refdata)
    num_points = mask.sum()
    if num_points == 0:
        mnmb = np.nan
        fge = np.nan
    else:
        tmp = difference[mask] / sum_data_refdata[mask]
        if weights is not None:
            weights = weights[mask]
        mnmb = 2.0 / num_points * sum(tmp, weights=weights)
        fge = 2.0 / num_points * sum(np.abs(tmp), weights=weights)

    result["nmb"] = nmb
    result["mnmb"] = mnmb
    result["fge"] = fge

    return result


def closest_index(num_array, value):
    """Returns index in number array that is closest to input value"""
    return np.argmin(np.abs(np.asarray(num_array) - value))


def numbers_in_str(input_string):
    """This method finds all numbers in a string

    Note
    ----
    - Beta version, please use with care
    - Detects only integer numbers, dots are ignored

    Parameters
    ----------
    input_string : str
        string containing numbers

    Returns
    -------
    list
        list of strings specifying all numbers detected in string

    Example
    -------
    >>> numbers_in_str('Bla42Blub100')
    [42, 100]
    """
    numbers = []
    IN_NUM = False
    c_num = None
    for char in input_string:
        try:
            int(char)
            if not IN_NUM:
                IN_NUM = True
                c_num = char
            elif IN_NUM:
                c_num += char
        except Exception:
            if IN_NUM:
                numbers.append(c_num)
            IN_NUM = False
    if IN_NUM:
        numbers.append(c_num)
    return numbers


def exponent(num):
    """Get exponent of input number

    Parameters
    ----------
    num : :obj:`float` or iterable
        input number

    Returns
    -------
    :obj:`int` or :obj:`ndarray` containing ints
        exponent of input number(s)

    Example
    -------
    >>> from pyaerocom.mathutils import exponent
    >>> exponent(2340)
    3
    """
    return np.floor(np.log10(abs(np.asarray(num)))).astype(int)


def range_magnitude(low, high):
    """Returns magnitude of value range

    Parameters
    ----------
    low : float
        lower end of range
    high : float
        upper end of range

    Returns
    -------
    int
        magnitudes spanned by input numbers

    Example
    -------

    >>> range_magnitude(0.1, 100)
    3
    >>> range_magnitude(100, 0.1)
    -3
    >>> range_magnitude(1e-3, 1e6)
    9

    """
    return exponent(high) - exponent(low)


def estimate_value_range(vmin, vmax, extend_percent=0):
    """
    Round and extend input range to estimate lower and upper bounds of range

    Parameters
    ----------
    vmin : float
        lower value of range
    vmax : float
        upper value of range
    extend_percent : int
        percentage specifying to which extent the input range is supposed to be
        extended.

    Returns
    -------
    float
        estimated lower end of range
    float
        estimated upper end of range


    """
    if not vmax > vmin:
        raise ValueError("vmax needs to exceed vmin")
    # extent value range by +/- 5%
    offs = (vmax - vmin) * extend_percent * 0.01
    vmin, vmax = vmin - offs, vmax + offs

    if vmin != 0:
        exp = float(exponent(vmin))
    else:
        exp = float(exponent(vmax))
    # round values
    vmin = np.floor(vmin * 10 ** (-exp)) * 10.0 ** (exp)
    vmax = np.ceil(vmax * 10 ** (-exp)) * 10.0 ** (exp)
    return vmin, vmax


def _init_stats_dummy():
    # dummy for statistics dictionary for locations without data
    stats_dummy = {}
    for k in calc_statistics([1], [1]):
        stats_dummy[k] = np.nan

    # Test to make sure these variables are defined even when yearly and season != all
    stats_dummy["R_spatial_mean"] = np.nan
    stats_dummy["R_spatial_median"] = np.nan
    stats_dummy["R_temporal_mean"] = np.nan
    stats_dummy["R_temporal_median"] = np.nan

    return stats_dummy



VAR_PREFIXES = ["vmr", "mmr", "conc", "sconc", "wet", "dry"]

# in g/mol
MOLMASSES = {
    "air_dry": 28.9647,
    "o3": 48,
    "so2": 64.066,
    "so4": 96.06,
    "no": 30.01,
    "no2": 46.0055,
    "hno3": 63.01,
    "nh3": 17.031,
    "co": 28.010,
    "isop": 68.12,
    "glyoxal": 58.036,
    "glyox": 58.036,
    "hcho": 30.026,
}


class UnkownSpeciesError(ValueError):
    pass


def get_species(var_name):
    """
    Get species name from variable name

    Parameters
    ----------
    var_name : str
        pyaerocom variable name (cf. variables.ini)

    Raises
    ------
    UnkownSpeciesError
        if species cannot be inferred

    Returns
    -------
    str
        name of species

    """
    if var_name in MOLMASSES:
        return var_name
    for prefix in VAR_PREFIXES:
        if var_name.startswith(prefix):
            species = var_name.split(prefix)[-1]
            if species in MOLMASSES:
                return species
    raise UnkownSpeciesError(f"Could not infer atom / molecule/ species from var_name {var_name}")


def get_molmass(var_name):
    """
    Get molar mass for input variable

    Parameters
    ----------
    var_name : str
        pyaerocom variable name (cf. variables.ini) or name of species

    Returns
    -------
    float
        molar mass of species in units of g/mol

    """
    return MOLMASSES[get_species(var_name)]


def get_mmr_to_vmr_fac(var_name):
    """
    Get conversion factor for MMR -> VMR conversion for input variable

    Note
    ----
    Assumes dry air molar mass

    Parameters
    ----------
    var_name : str
        Name of variable to be converted

    Returns
    -------
    float
        multiplication factor to convert MMR -> VMR

    """
    return get_molmass("air_dry") / get_molmass(var_name)



"""
read Aeronet SDA V2 data
"""
import os
import re

import numpy as np
import pandas as pd

from pyaerocom import const as const
from pyaerocom.aux_var_helpers import (
    calc_ang4487aer,
    calc_od550aer,
    calc_od550gt1aer,
    calc_od550lt1aer,
)
from pyaerocom.io.readaeronetbase import ReadAeronetBase
from pyaerocom.stationdata import StationData


class ReadAeronetSdaV2(ReadAeronetBase):
    """Interface for reading Aeronet Sun V2 Level 2 data

    Note
    ----
    Maintenance for AERONET version 2 data is deprecated since pyaerocom
    version 0.11.0. Please use version 3 data and associated pyaerocom readers.
    """

    #: Mask for identifying datafiles
    _FILEMASK = "*.ONEILL_20"

    #: version log of this class (for caching)
    __version__ = "0.08_" + ReadAeronetBase.__baseversion__

    #: Name of dataset (OBS_ID)
    DATA_ID = const.AERONET_SUN_V2L2_SDA_DAILY_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [const.AERONET_SUN_V2L2_SDA_DAILY_NAME]

    #: dictionary assigning temporal resolution flags for supported datasets
    #: that are provided in a defined temporal resolution
    TS_TYPES = {const.AERONET_SUN_V2L2_SDA_DAILY_NAME: "daily"}

    #: default variables for read method
    DEFAULT_VARS = ["od550aer", "od550gt1aer", "od550lt1aer"]

    #: value corresponding to invalid measurement
    NAN_VAL = -9999.0

    # Comment jgliss 20180719: translated most varname to Aerocom convention.
    # Those, for which no defined name is known were named with a leading
    # underscore (e.g. _eta500lt1 for fine mode fraction at 500nm)
    #: Dictionary that specifies the index for each data column
    COL_INDEX = dict(
        date=0,  # Date(dd:mm:yyyy)
        time=1,  # Time(hh:mm:ss),
        julien_day=2,  # Julian_Day
        od500aer=3,  # Total_AOD_500nm[tau_a]
        od500lt1aer=4,  # Fine_Mode_AOD_500nm[tau_f]
        od500gt1aer=5,  # Coarse_Mode_AOD_500nm[tau_c]
        _eta500lt1=6,  # FineModeFraction_500nm[eta]
        _aod500aer_fiterr=7,  # 2nd_Order_Reg_Fit_Error-Total_AOD_500nm[regression_dtau_a]
        _aod500lt1aer_rmse=8,  # RMSE_Fine_Mode_AOD_500nm[Dtau_f]
        _aod500gt1aer_rmse=9,  # RMSE_Coarse_Mode_AOD_500nm[Dtau_c]
        _eta500lt1_rmse=10,  # RMSE_FineModeFraction_500nm[Deta]
        _ang50aer=11,  # Angstrom_Exponent(AE)-Total_500nm[alpha]
        od870aer=15,  # 870nm_Input_AOD
        od675aer=16,  # 675nm_Input_AOD
        od667aer=17,  # 667nm_Input_AOD
        od555aer=18,  # 555nm_Input_AOD
        od551aer=19,  # 551nm_Input_AOD
        od532aer=20,  # 532nm_Input_AOD
        od531aer=21,  # 531nm_Input_AOD
        od500aer_input=22,  # 500nm_Input_AOD
        od490aer=23,  # 490nm_Input_AOD
        od443aer=24,  # 443nm_Input_AOD
        od440aer=25,  # 440nm_Input_AOD
        od412aer=26,  # 412nm_Input_AOD
        od380aer=27,
    )  # 380nm_Input_AOD

    # FURTHER HEADER COLUMN INDICES

    # i_dAE/dln(wavelength)-Total_500nm[alphap]=12,
    # i_AE-Fine_Mode_500nm[alpha_f]=13,
    # i_dAE/dln(wavelength)-Fine_Mode_500nm[alphap_f]=14,
    # LAST_PROCESSING_DATE_INDEX = 28, #Last_Processing_Date
    # NUMBER_OF_WAVELENGTHS_INDEX = 29, # Number_of_Wavelengths
    # UMBER_OF_OBSERVATIONS_INDEX = 30, #Number_of_Observations
    # EXACT_WAVELENGTHS_FOR_INPUT_AOD_INDEX = 31) # Exact_Wavelengths_for_Input_AOD

    # specify required dependencies for auxiliary variables, i.e. variables
    # that are NOT in Aeronet files but are computed within this class.
    # For instance, the computation of the AOD at 550nm requires import of
    # the AODs at 440, 500 and 870 nm.

    #: dictionary containing information about additionally required variables
    #: for each auxiliary variable (i.e. each variable that is not provided
    #: by the original data but computed on import)
    AUX_REQUIRES = {
        "ang4487aer": ["od440aer", "od870aer"],
        "od550aer": ["od500aer", "ang4487aer"],
        "od550gt1aer": ["od500gt1aer", "ang4487aer"],
        "od550lt1aer": ["od500lt1aer", "ang4487aer"],
    }

    #: Functions that are used to compute additional variables (i.e. one
    #: for each variable defined in AUX_REQUIRES)
    AUX_FUNS = {
        "ang4487aer": calc_ang4487aer,
        "od550aer": calc_od550aer,
        "od550gt1aer": calc_od550gt1aer,
        "od550lt1aer": calc_od550lt1aer,
    }

    #: List of variables that are provided by this dataset (will be extended
    #: by auxiliary variables on class init, for details see __init__ method of
    #: base class ReadUngriddedBase)
    PROVIDES_VARIABLES = list(COL_INDEX)

    # This how the beginning of data file looks like

    # Level 2.0 Quality Assured Data. The following AERONET-SDA data are
    # derived from AOD data which are pre and post-field calibrated and
    # manually inspected. SDA Version 4.1 (tauf_tauc),Note: the labels in
    # square brackets that follow some of the parameter (column) names are
    # the symbols associated with these parameters in the original SDA
    # publication of O'Neill et al. (2003)
    # Location=Zvenigorod,Latitude=55.695000,Longitude=36.775000,
    # Elevation[m]=200.000000,PI=Brent_Holben,Email=Brent.N.Holben@nasa.gov
    # SDA from Level 2.0 AOD,Daily Average,UNITS can be found at
    # http://aeronet.gsfc.nasa.gov/data_menu.html
    # Date(dd:mm:yyyy),
    # Time(hh:mm:ss),
    # Julian_Day,Total_AOD_500nm[tau_a],Fine_Mode_AOD_500nm[tau_f],
    # Coarse_Mode_AOD_500nm[tau_c],FineModeFraction_500nm[eta],
    # 2nd_Order_Reg_Fit_Error-Total_AOD_500nm[regression_dtau_a],
    # RMSE_Fine_Mode_AOD_500nm[Dtau_f],RMSE_Coarse_Mode_AOD_500nm[Dtau_c],
    # RMSE_FineModeFraction_500nm[Deta],Angstrom_Exponent(AE)-
    # Total_500nm[alpha],dAE/dln(wavelength)-Total_500nm[alphap],
    # AE-Fine_Mode_500nm[alpha_f],
    # dAE/dln(wavelength)-Fine_Mode_500nm[alphap_f],
    # 870nm_Input_AOD,675nm_Input_AOD,667nm_Input_AOD,555nm_Input_AOD,
    # 551nm_Input_AOD,532nm_Input_AOD,531nm_Input_AOD,
    # 500nm_Input_AOD,490nm_Input_AOD,443nm_Input_AOD,440nm_Input_AOD,
    # 412nm_Input_AOD,380nm_Input_AOD,Last_Processing_Date,
    # Number_of_Wavelengths,Number_of_Observations,
    # Exact_Wavelengths_for_Input_AOD
    # 16:09:2006,00:00:00,259.000000,0.059239,0.032410,0.026828,0.571506,
    # 0.004645,0.007370,0.005064,0.091573,1.278830,-0.656569,2.354811,
    # 1.413517,0.036734,0.039337,-999.,-999.,-999.,-999.,-999.,0.064670,
    # -999.,-999.,0.069614,-999.,0.083549,27:11:2007,5,11,0.868800,
    # 0.675600,0.440400,0.500500,0.380100

    @property
    def col_index(self):
        """Dictionary that specifies the index for each data column

        Note
        ----
        Pointer to :attr:`COL_INDEX`
        """
        return self.COL_INDEX

    def read_file(self, filename, vars_to_retrieve=None, vars_as_series=False):
        """Read Aeronet Sun SDA V2 file

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : :obj:`list`, optional
            list of str with variable names to read. If None, use
            :attr:`DEFAULT_VARS`
        vars_as_series : bool
            if True, the data columns of all variables in the result dictionary
            are converted into pandas Series objects

        Returns
        -------
        StationData
            dict-like object containing results
        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        # implemented in base class
        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(vars_to_retrieve)

        # create empty data object (is dictionary with extended functionality)
        data_out = StationData()

        data_out.data_id = self.data_id
        # create empty arrays for meta information
        for item in self.META_NAMES_FILE:
            data_out[item] = []

        # create empty arrays for all variables that are supposed to be read
        # from file
        for var in vars_to_read:
            data_out[var] = []

        # Iterate over the lines of the file
        self.logger.info(f"Reading file {filename}")
        # Iterate over the lines of the file
        with open(filename) as in_file:

            c_head_line = in_file.readline()
            c_algorithm = in_file.readline()

            self.logger.info(c_head_line)
            self.logger.info(c_algorithm)

            c_dummy = in_file.readline()
            # re.split(r'=|\,',c_dummy)
            i_dummy = iter(re.split(r"=|\,", c_dummy.rstrip()))
            dict_loc = dict(zip(i_dummy, i_dummy))

            data_out["latitude"] = float(dict_loc["Latitude"])
            data_out["longitude"] = float(dict_loc["Longitude"])
            data_out["altitude"] = float(dict_loc["Elevation[m]"])
            data_out["station_name"] = dict_loc["Location"]
            data_out["PI"] = dict_loc["PI"]
            data_out["ts_type"] = self.TS_TYPE

            c_dummy = in_file.readline()
            c_Header = in_file.readline()

            self.logger.info(c_Header)

            col_index = self.col_index

            vars_available = {}
            for var in vars_to_read:
                if var in col_index:
                    vars_available[var] = col_index[var]
                else:
                    self.logger.warning(
                        f"Variable {var} not available in file {os.path.basename(filename)}"
                    )

            for line in in_file:
                # process line
                dummy_arr = line.split(self.COL_DELIM)

                # This uses the numpy datestring64 functions that e.g. also
                # support Months as a time step for timedelta
                # Build a proper ISO 8601 UTC date string
                day, month, year = dummy_arr[col_index["date"]].split(":")
                # pdb.set_trace()
                datestring = "-".join([year, month, day])
                datestring = "T".join([datestring, dummy_arr[col_index["time"]]])
                # NOTE JGLISS: parsing timezone offset was removed on 22/2/19
                # since it is deprecated in recent numpy versions, for details
                # see https://www.numpy.org/devdocs/reference/arrays.datetime.html#changes-with-numpy-1-11
                # datestring = '+'.join([datestring, '00:00'])
                data_out["dtime"].append(np.datetime64(datestring))

                # copy the data fields
                for var, idx in vars_available.items():
                    val = np.float_(dummy_arr[idx])
                    if val == self.NAN_VAL:
                        val = np.nan
                    data_out[var].append(val)

        # convert all lists to numpy arrays
        data_out["dtime"] = np.asarray(data_out["dtime"])

        for var in vars_to_read:
            if var in vars_available:
                array = np.asarray(data_out[var])
            else:
                array = np.zeros(len(data_out["dtime"])) * np.nan
            data_out[var] = array

        # compute additional variables (if applicable)
        data_out = self.compute_additional_vars(data_out, vars_to_compute)

        # convert data vectors to pandas.Series (if applicable)
        if vars_as_series:
            for var in vars_to_read + vars_to_compute:
                if var in vars_to_retrieve:
                    data_out[var] = pd.Series(data_out[var], index=data_out["dtime"])
                else:
                    del data_out[var]

        return data_out



#: Default variable ranges for web display
var_ranges_defaults = {
    "default": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "ang4487aer": {"scale": [0, 0.25, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2], "colmap": "coolwarm"},
    "od550aer": {
        "scale": [0, 0.05, 0.1, 0.15, 0.20, 0.25, 0.30, 0.35, 0.40],
        "colmap": "coolwarm",
    },
    "od550lt1aer": {
        "scale": [0, 0.05, 0.1, 0.15, 0.20, 0.25, 0.30, 0.35, 0.40],
        "colmap": "coolwarm",
    },
    "od550gt1aer": {
        "scale": [0, 0.05, 0.1, 0.15, 0.20, 0.25, 0.30, 0.35, 0.40],
        "colmap": "coolwarm",
    },
    "abs550aer": {
        "scale": [0, 0.0125, 0.025, 0.0375, 0.05, 0.0625, 0.075, 0.0875, 0.1],
        "colmap": "coolwarm",
    },
    "absc550aer": {"scale": [0, 12.5, 25, 37.5, 50, 62.5, 75, 87.5, 100], "colmap": "coolwarm"},
    "scatc550dryaer": {
        "scale": [0, 12.5, 25, 37.5, 50, 62.5, 75, 87.5, 100],
        "colmap": "coolwarm",
    },
    "extinction": {
        "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
        "colmap": "coolwarm",
    },
    "backscatter": {
        "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
        "colmap": "coolwarm",
    },
    "concso4": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "concso2": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "concpm10": {"scale": [0, 10, 20, 30, 40, 50, 60, 70, 80], "colmap": "coolwarm"},
    "concpm25": {"scale": [0, 10, 20, 30, 40, 50, 60, 70, 80], "colmap": "coolwarm"},
    "conco3": {"scale": [0, 15, 30, 45, 60, 75, 90, 105, 120], "colmap": "coolwarm"},
    "vmro3": {"scale": [0, 7.5, 15, 22.5, 30, 37.5, 45, 52.5, 60], "colmap": "coolwarm"},
    "concno2": {"scale": [0, 10, 20, 30, 40, 50, 60, 70, 80], "colmap": "coolwarm"},
    "vmrno2": {"scale": [0, 5, 10, 15, 20, 25, 30, 35, 40], "colmap": "coolwarm"},
    "concNhno3": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "concNno3pm10": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "concNno3pm25": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "concNnh3": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "concNnh4": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "concNtno3": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "concNtnh": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "concsspm25": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "concsspm10": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "concCecpm25": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "concCocpm25": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "wetoxs": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "wetoxn": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "wetrdn": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "dryoxs": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "dryoxn": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
    "dryrdn": {"scale": [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10], "colmap": "coolwarm"},
}

#: Default information for statistical parameters
statistics_defaults = {
    "nmb": {
        "name": "NMB",
        "longname": "Normalized Mean Bias",
        "scale": [-100, -75, -50, -25, 0, 25, 50, 75, 100],
        "colmap": "bwr",
        "unit": "%",
        "decimals": 1,
    },
    "mnmb": {
        "name": "MNMB",
        "longname": "Modified Normalized Mean Bias",
        "scale": [-100, -75, -50, -25, 0, 25, 50, 75, 100],
        "colmap": "bwr",
        "unit": "%",
        "decimals": 1,
    },
    "R": {
        "name": "R",
        "longname": "Correlation Coefficient",
        "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
        "colmap": "RdYlGn",
        "unit": "1",
        "decimals": 2,
    },
    "R_spearman": {
        "name": "R Spearman",
        "longname": "R Spearman Correlation",
        "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
        "colmap": "RdYlGn",
        "unit": "1",
        "decimals": 2,
    },
    "fge": {
        "name": "FGE",
        "longname": "Fractional Gross Error",
        "scale": [0, 0.25, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2],
        "colmap": "reverseColmap(RdYlGn)",
        "unit": "1",
        "decimals": 2,
    },
    "nrms": {
        "name": "NRMSE",
        "longname": "Normalized Root Mean Square Error",
        "scale": [0, 25, 50, 75, 100, 125, 150, 175, 200],
        "colmap": "Reds",
        "unit": "%",
        "decimals": 1,
    },
    "rms": {
        "name": "RMSE",
        "longname": "Root Mean Square Error",
        "scale": None,
        "colmap": "coolwarm",
        "unit": "1",
        "decimals": 2,
    },
    "data_mean": {
        "name": "Mean-Mod",
        "longname": "Model Mean",
        "scale": None,
        "colmap": "coolwarm",
        "unit": "1",
        "decimals": 2,
    },
    "refdata_mean": {
        "name": "Mean-Obs",
        "longname": "Observation Mean",
        "scale": None,
        "colmap": "coolwarm",
        "unit": "1",
        "decimals": 2,
    },
}

# Default information for additional statistical parameters
extended_statistics = {
    "R_spatial_mean": {
        "name": "R-Space",
        "longname": "Spatial R computed from yearly averages",
        "overall_only": True,
        "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
        "colmap": "RdYlGn",
        "unit": "1",
        "decimals": 2,
    },
    "R_temporal_median": {
        "name": "R-Temporal",
        "longname": "R temporal median",
        "overall_only": True,
        "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
        "colmap": "RdYlGn",
        "unit": "1",
        "decimals": 2,
    },
}

#: Default information about trend display
statistics_trend = {
    "obs/mod_trend": {
        "name": "Obs/Mod-Trends",
        "longname": "Trends",
        "scale": [-10.0, -7.5, -5.0, -2.5, 0, 2.5, 5.0, 7.5, 10.0],
        "colmap": "bwr",
        "unit": "%/yr",
        "decimals": 1,
    },
    "obs_trend": {
        "name": "Obs-Trends",
        "longname": "Observed Trends",
        "scale": [-10.0, -7.5, -5.0, -2.5, 0, 2.5, 5.0, 7.5, 10.0],
        "colmap": "bwr",
        "unit": "%/yr",
        "decimals": 1,
    },
    "mod_trend": {
        "name": "Mod-Trends",
        "longname": "Modelled Trends",
        "scale": [-10, -7.5, -5.0, -2.5, 0, 2.5, 5.0, 7.5, 10.0],
        "colmap": "bwr",
        "unit": "%/yr",
        "decimals": 1,
    },
}

#: Mapping of pyaerocom variable names to web naming conventions
var_web_info = dict(
    od550aer=["AOD", "2D", "Optical properties"],
    od550csaer=["AOD (clear sky)", "2D", "Optical properties"],
    od550lt1aer=["AODf", "2D", "Optical properties"],
    od550gt1aer=["AODc", "2D", "Optical properties"],
    od550dust=["AODdust", "2D", "Optical properties"],
    abs550aer=["AAOD", "2D", "Optical properties"],
    ang4487aer=["AE", "2D", "Optical properties"],
    angabs4487aer=["AAE", "2D", "Optical properties"],
    ang4487csaer=["AE (clear sky)", "2D", "Optical properties"],
    sc550dryaer=["Scat. coef. (dry)", "3D", "Optical properties"],
    sc550aer=["Scat. coef.", "3D", "Optical properties"],
    ac550aer=["Abs. coef.", "3D", "Optical properties"],
    ac550dryaer=["Abs. coef. (dry)", "3D", "Optical properties"],
    ec532aer=["Ext. coeff.", "3D", "Optical properties"],
    bsc532aer=["Backscat. coeff.", "3D", "Optical properties"],
    concpm10=["PM10", "3D", "Particle concentrations"],
    concpm25=["PM2.5", "3D", "Particle concentrations"],
    concso4=["SO4", "3D", "Particle concentrations"],
    concso4pr=["SO4 (precip.)", "3D", "Particle concentrations"],
    concbc=["BC", "3D", "Particle concentrations"],
    concoa=["OA", "3D", "Particle concentrations"],
    concss=["SS", "3D", "Particle concentrations"],
    conco3=["O3", "3D", "Gas concentrations"],
    concso2=["SO2", "3D", "Gas concentrations"],
    vmrox=["OX", "3D", "Gas volume mixing ratio"],
    vmrso2=["SO2", "3D", "Gas volume mixing ratio"],
    conctno3=["tNO3", "3D", "Concentration"],
    conctnh=["tNH", "3D", "Concentration"],
    concnh3=["NH3", "3D", "Concentration"],
    conchno3=["HNO3", "3D", "Concentration"],
    concno310=["NO3_PM10", "3D", "Particle concentration"],
    concno325=["NO3_PM25", "3D", "Particle concentration"],
    concss10=["SS_PM10", "3D", "Particle concentration"],
    concss25=["SS_PM25", "3D", "Particle concentration"],
    concec=["EC", "3D", "Particle concentration"],
    conccoc=["OC", "3D", "Particle concentration"],
    wetoxs=["WetOXS", "3D", "Deposition"],
    drysox=["DryOXS", "3D", "Deposition"],
    concnh4=["NH4", "3D", "Gas concentrations"],
    concno3=["NO3", "3D", "Gas concentrations"],
    vmro3=["O3", "3D", "Volume mixing ratios"],
    vmro3max=["O3Max", "3D", "Volume mixing ratios"],
    concNtno3=["tNO3", "3D", "Concentration"],
    concNtnh=["tNH", "3D", "Concentration"],
    concNnh3=["NH3", "3D", "Concentration"],
    concNhno3=["HNO3", "3D", "Concentration"],
    concNno3pm10=["NO3_PM10", "3D", "Particle concentration"],
    concNno3pm25=["NO3_PM25", "3D", "Particle concentration"],
    concsspm10=["SS_PM10", "3D", "Particle concentration"],
    concsspm25=["SS_PM25", "3D", "Particle concentration"],
    concCecpm25=["EC", "3D", "Particle concentration"],
    concCocpm25=["OC", "3D", "Particle concentration"],
    wetoxn=["WetOXN", "3D", "Deposition"],
    wetrdn=["WetRDN", "3D", "Deposition"],
    dryoxs=["FakeDryOXS", "3D", "Deposition"],
    dryoxn=["FakeDryOXN", "3D", "Deposition"],
    dryrdn=["FakeDryRDN", "3D", "Deposition"],
    concno2=["NO2", "3D", "Gas concentrations"],
    concNnh4=["NH4", "3D", "Gas concentrations"],
    vmrco=["CO", "3D", "Volume mixing ratios"],
    vmrno2=["NO2", "3D", "Volume mixing ratios"],
    concco=["CO", "3D", "Particle concentration"],
)



"""
Definitions and helpers related to time conversion
"""
from datetime import datetime

from iris import coord_categorisation

TS_TYPES = ["minutely", "hourly", "daily", "weekly", "monthly", "yearly", "native"]

# The following import was removed and the information about available unit
# strings was copied from the netCDF4 module directly here
microsec_units = ["microseconds", "microsecond", "microsec", "microsecs"]
millisec_units = ["milliseconds", "millisecond", "millisec", "millisecs"]
sec_units = ["second", "seconds", "sec", "secs", "s"]
min_units = ["minute", "minutes", "min", "mins"]
hr_units = ["hour", "hours", "hr", "hrs", "h"]
day_units = ["day", "days", "d"]

#
# Start of the gregorian calendar
# adapted from here: https://github.com/Unidata/cftime/blob/master/cftime/_cftime.pyx
GREGORIAN_BASE = datetime(1582, 10, 15)

IRIS_AGGREGATORS = {
    "hourly": coord_categorisation.add_hour,
    "daily": coord_categorisation.add_day_of_year,
    "monthly": coord_categorisation.add_month_number,
    "yearly": coord_categorisation.add_year,
}

# some helper dictionaries for conversion of temporal resolution
# https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
TS_TYPE_TO_PANDAS_FREQ = {
    "minutely": "T",
    "hourly": "H",
    "daily": "D",
    "weekly": "W-MON",
    "monthly": "MS",  # Month start !
    "season": "Q",
    "yearly": "AS",
}

PANDAS_RESAMPLE_OFFSETS = {"AS": "181D", "MS": "14D"}

PANDAS_FREQ_TO_TS_TYPE = {v: k for k, v in TS_TYPE_TO_PANDAS_FREQ.items()}

# frequency strings
# https://numpy.org/devdocs/reference/arrays.datetime.html#datetime-units
TS_TYPE_TO_NUMPY_FREQ = {
    "minutely": "m",
    "hourly": "h",
    "daily": "D",
    "weekly": "W",
    "monthly": "M",  # Month start !
    "yearly": "Y",
}

# conversion of ts_types to strings that cf_units understands
TS_TYPE_TO_SI = {
    "minutely": "min",
    "hourly": "h",
    "daily": "d",
    "weekly": "week",
    "monthly": "month",
    "yearly": "yr",
}

SI_TO_TS_TYPE = {v: k for k, v in TS_TYPE_TO_SI.items()}

# conversion of datetime-like objects for given temporal resolutions (can, e.g.
# be used in plotting methods)
TS_TYPE_DATETIME_CONV = {
    None: "%d.%m.%Y",  # Default
    "hourly": "%d.%m.%Y",
    "3hourly": "%d.%m.%Y",
    "daily": "%d.%m.%Y",
    "weekly": "%d.%m.%Y",
    "monthly": "%b %Y",
    "yearly": "%Y",
}

TS_TYPE_SECS = {
    "minutely": 60,
    "hourly": 3600,
    "3hourly": 10800,
    "daily": 86400,
    "weekly": 604800,
    "monthly": 2592000,  # counting 3 days per month (APPROX)
    "yearly": 31536000,  # counting 365 days (APPROX)
}



from . import config, heatmaps, mapping, plotcoordinates, plotscatter
from .heatmaps import df_to_heatmap
from .mapping import init_map, plot_griddeddata_on_map
from .plotcoordinates import plot_coordinates
from .plotscatter import plot_scatter
